""" Utilities for constructing keytabs data using the attributes of a GssKerberosKey and combining keytabs.

See this page for more info on the evolution of the format beyond the code comments explaining how things work
conceptually: http://manpages.ubuntu.com/manpages/artful/man3/krb5_fileformats.3.html
"""
import binascii
import math
import time

from ms_active_directory import logging_utils

from typing import List


from ms_active_directory.environment.kerberos.kerberos_constants import (
    DEFAULT_UNKNOWN_NAME_TYPE,
    ENCRYPTION_TYPE_FIELD_SIZE,
    ENTRY_LENGTH_FIELD_SIZE_BYTES,
    FLAGS_FIELD_SIZE_BYTES,
    KEYTAB_FORMAT_SIZE_BYTES,
    KEYTAB_FORMAT_VERSION_FOR_KEYTAB_FORMAT_VERSION,
    KEY_LENGTH_FIELD_SIZE_BYTES,
    KRB5_ENC_TYPE_TO_ENC_TYPE_VALUE_MAP,
    LEADING_HEX_FOR_KERBEROS_V5,
    NUM_COMPONENTS_FIELD_SIZE_BYTES,
    REALM_LENGTH_FIELD_SIZE_BYTES,
    PREFERRED_KEYTAB_FORMAT_VERSION,
    PRINCIPAL_COMPONENT_DIVIDER,
    PRINCIPAL_COMPONENT_LENGTH_FIELD_SIZE_BYTES,
    PRINCIPAL_TYPE_FIELD_SIZE_BYTES,
    TIMESTAMP_FIELD_SIZE_BYTES,
    VNO8_FIELD_SIZE_BYTES,
    VNO32_FIELD_SIZE_BYTES,
)
from ms_active_directory.exceptions import KeytabEncodingException


logger = logging_utils.get_logger()


def write_gss_kerberos_key_list_to_raw_bytes(gss_key_list: List, keytab_format_version: int=None):
    """ Given a list of fully formed keytab dictionaries, convert them to bytes representing all of them as a single
    keytab.
    This just builds upon the function to convert keytabs to hex, and then transforms the hex to bytes.
    """
    hex_data = write_gss_kerberos_key_list_to_raw_hex(gss_key_list, keytab_format_version)
    return binascii.unhexlify(hex_data)


def write_gss_kerberos_key_list_to_raw_hex(gss_key_list: List, keytab_format_version: int=None):
    """ Given a list of GssKerberosKey objects, convert them to a hex string representing all of them as a
    single kerberos keytab, usable for standard GSS authentication (either initiating or receiving).
    Optionally, a format version may be specified. If a format version is not provided, then all keytabs will be
    written in the format of the first keytab provided. If the first keytab provided does not have a format version
    then we will default to our preferred keytab format version.

    :returns: A hex-encoded string representing the data of a fully-formed keytab file, with no leading 0x in order
              to allow for easier writing out to a file/translation to bytes.
    """
    # if we have no keys just return
    if len(gss_key_list) == 0:
        return ''

    leading_keytab = gss_key_list.pop(0)
    # if no keytab format version is specified, assume that all of them should be written in the same format as the
    # first one
    if keytab_format_version is None:
        keytab_format_version = leading_keytab.format_version
    if keytab_format_version is None:
        keytab_format_version = PREFERRED_KEYTAB_FORMAT_VERSION

    keytab_hex_string = write_complete_keytab_structure_to_raw_hex_data(leading_keytab,
                                                                        keytab_format_version=keytab_format_version)
    # process remaining keys (if any)
    for keytab_dict in gss_key_list:
        keytab_hex_string += write_complete_keytab_structure_to_raw_hex_data(keytab_dict,
                                                                             keytab_format_version=keytab_format_version,
                                                                             include_keytab_format_prefix=False)
    return keytab_hex_string


def write_complete_keytab_structure_to_raw_bytes(gss_key, keytab_format_version: int=None,
                                                 include_keytab_format_prefix: bool=True,
                                                 use_current_time_for_keytab: bool=False):
    """ This is just a wrapper around write_complete_keytab_structure_to_raw_hex_data - go read its docstring
    This takes the result of write_complete_keytab_structure_to_raw_hex_data and encodes it to bytes before
    returning.
    """
    hex_data = write_complete_keytab_structure_to_raw_hex_data(gss_key, keytab_format_version,
                                                               include_keytab_format_prefix,
                                                               use_current_time_for_keytab)
    return binascii.unhexlify(hex_data)


def write_complete_keytab_structure_to_raw_hex_data(gss_key, keytab_format_version: int=None,
                                                    include_keytab_format_prefix: bool=True,
                                                    use_current_time_for_keytab: bool=False):
    """ This function, given a GssKerberosKey, will write it to bytes.
    Optionally, it can be forced into a specific keytab format version. If that isn't specified, then it will be
    written in the format version in the GssKerberosKey object.
    Optionally, the keytab format prefix and leading byte to indicate kerberos version 5 can be excluded. This is
    particularly useful if you plan to combine the result of this with other keytabs in a merge. But by default the
    byte is included, so the return value is a valid standalone keytab that can be used for authentication without
    additional processing.

    This function is essentially the behavior of the kerberos_keytab_ingesterr.py file in reverse.
    :param gss_key: The GssKerberosKey to turn into keytab data
    :param keytab_format_version: The format version to use for encoding numbers in the keytab
    :param include_keytab_format_prefix: If true, the leading 0x5 byte is included, along with the keytab format version
                                         written to a number, at the start of the keytab. This makes it ready for use
                                         in authentication.
    :param use_current_time_for_keytab: If true, overwrite the timestamp in the gss_key with the current time. This is
                                        the behavior of many command line kerberos tools like ktutil. If false, the
                                        timestamp is preserved. If no timestamp exists for the gss_key, then the current
                                        time will always be used. Defaults to False because I think the behavior of
                                        ktutil is stupid, and having a keytab's data change whenever I copy it from
                                        one file to another is confusing.
    """
    if keytab_format_version is None:
        keytab_format_version = gss_key.format_version

    keytab_hex_lead = ''

    if include_keytab_format_prefix:
        keytab_hex_lead = LEADING_HEX_FOR_KERBEROS_V5
        keytab_hex_lead += _write_number_to_hex(keytab_format_version,
                                                number_repr_size_in_bytes=KEYTAB_FORMAT_SIZE_BYTES,
                                                keytab_format_version=KEYTAB_FORMAT_VERSION_FOR_KEYTAB_FORMAT_VERSION)

    # track our keytab body separate from the lead so we don't include the lead in entry length analyses
    keytab_hex_body = ''

    # the first thing we'll encode is the number of components. this is recalculated by breaking up the gss_key's
    # principal
    num_components = len(gss_key.principal.split(PRINCIPAL_COMPONENT_DIVIDER))
    # if our keytab format version is 1, then the number of components is supposed to be 1 too large because there
    # was literally a bug in the v1 format in how it accounted for components :(
    if keytab_format_version == 1:
        num_components += 1
    keytab_hex_body += _write_number_to_hex(num_components, number_repr_size_in_bytes=NUM_COMPONENTS_FIELD_SIZE_BYTES,
                                            keytab_format_version=keytab_format_version)

    # next, we write the realm
    keytab_hex_body += _write_realm_to_hex(gss_key.realm, keytab_format_version=keytab_format_version)

    # third we write our principal
    keytab_hex_body += _write_principal_to_hex(gss_key.principal, keytab_format_version=keytab_format_version)

    # the last piece of information about the principal to write is the principal type. this doesn't functionally matter
    # and many KDCs will populate a generic GSS API name type here.
    # However, this information can be helpful for figuring out the intention of the KDC when the keytab was first
    # created. For example, users and servers are different types, so if the manager intended for this to the user
    # keytab OF A SERVER (to let the server query for info about users it's authenticating/authorizing) then this info
    # can help differentiate it from true user keys
    # keytab format version 1 does not include this value
    name_type = gss_key.gss_name_type if gss_key.gss_name_type is not None else DEFAULT_UNKNOWN_NAME_TYPE
    if keytab_format_version != 1:
        keytab_hex_body += _write_number_to_hex(name_type, PRINCIPAL_TYPE_FIELD_SIZE_BYTES,
                                                keytab_format_version=keytab_format_version)

    # the next two things to write are the time of keytab generation and the old 8-bit version of kvno.
    # the 8-bit version is useless everywhere except for old machines (like ubuntu 14 with an older libkrb5) because
    # we'll also encode the 32-bit version which will be preferred
    # the time is just the time the file was generated, and so it changes anytime keytabs are passed through ktutil.
    # this is stupid
    # if the caller indicated to use the current time rather than the time in the gss_key, or if the gss_key has no
    # timestamp, then use the current time
    current_time = int(time.time())
    timestamp = gss_key.timestamp
    if use_current_time_for_keytab or gss_key.timestamp is None:
        timestamp = current_time
    keytab_hex_body += _write_number_to_hex(timestamp, TIMESTAMP_FIELD_SIZE_BYTES,
                                            keytab_format_version=keytab_format_version)

    # according to spec, if our kvno doesn't fit in 8 bits (>255), then we write the lower 8 bits of our kvno for the
    # 8-bit version of kvno. this is ignored in favor of the 32 bit field
    kvno32 = gss_key.kvno
    kvno8 = kvno32
    if kvno32 > 2**8-1:
        bit_string = bin(kvno32)
        last_8 = bit_string[-8:]
        kvno8 = int(last_8, 2)
    keytab_hex_body += _write_number_to_hex(kvno8, VNO8_FIELD_SIZE_BYTES,
                                            keytab_format_version=keytab_format_version)

    # next we write the encryption type and the secret key itself.
    # the encryption type tells us how our key was formed and how it should be used as a shared secret. many keys will
    # be the same size, and a good encryption key doesn't betray what encryption scheme was used to create it by its
    # appearance, so we encode encryption type so that we can understand how to encrypt/sign/decrypt data
    # our key is the only part of the keytab entry that ACTUALLY matters cryptographically for authentication, signing,
    # and sealing with kerberos.
    # encryption type comes first
    enc_type = gss_key.encryption_type
    enc_type_value = KRB5_ENC_TYPE_TO_ENC_TYPE_VALUE_MAP.get(enc_type)
    keytab_hex_body += _write_number_to_hex(enc_type_value, ENCRYPTION_TYPE_FIELD_SIZE,
                                            keytab_format_version=keytab_format_version)
    keytab_hex_body += _write_raw_kerberos_key_to_keytab_entry(gss_key.raw_key,
                                                               keytab_format_version=keytab_format_version)

    # now write the 32-bit version of our key version number
    keytab_hex_body += _write_number_to_hex(kvno32, VNO32_FIELD_SIZE_BYTES, keytab_format_version=keytab_format_version)

    # flags are optional. only encode them if they're non-null
    flags = gss_key.flags
    if flags is not None:  # flags could be 0 as it's numeric
        keytab_hex_body += _write_number_to_hex(flags, FLAGS_FIELD_SIZE_BYTES,
                                                keytab_format_version=keytab_format_version)

    # so calculate the size of the entry now that we've encoded all of the data and prepend it
    entry_length = len(keytab_hex_body)
    # 2 hex digits = 1 byte, so our byte length is half our string length since we used hex
    entry_length_bytes = entry_length//2
    entry_length_field = _write_number_to_hex(entry_length_bytes,
                                              number_repr_size_in_bytes=ENTRY_LENGTH_FIELD_SIZE_BYTES,
                                              keytab_format_version=keytab_format_version)
    keytab_hex_body = entry_length_field + keytab_hex_body

    return keytab_hex_lead + keytab_hex_body


def _write_number_to_hex(number_to_write: int, number_repr_size_in_bytes: int, keytab_format_version: int):
    """ Write any number to hex, while allocating the number of bytes specified.
    Number fields in a keytab occupy a fixed size, so we make sure to account for that when encoding them to hex.
    """
    byte_order = 'little'
    if keytab_format_version == 2:
        byte_order = 'big'
    elif keytab_format_version != 1:
        raise KeytabEncodingException('Invalid keytab format version {}. Format version must be 1 or 2'
                                      .format(keytab_format_version))
    # this raises an exception if the number to write is too big to fit in number_repr_size_in_bytes
    return number_to_write.to_bytes(number_repr_size_in_bytes, byte_order).hex()


def _write_string_to_hex(string_to_write: str):
    """ Write any string to hex.
    As mentioned above, numbers get padded because all numbers are a fixed size in keytabs.
    However, strings are super free-form, like principals and realms. They're not constrained to a fixed size ever, and
    so instead all string fields will also end up encoding their length before them in the keytab. So there's no need
    for any input other than the string itself to this function.
    """
    return string_to_write.encode().hex()


def _write_string_and_its_length_to_hex(any_string: str, size_of_length_field_bytes: int, keytab_format_version: int):
    """ Write any string, and the information needed to read it back from a keytab, to hex.
    This is done backwards - all strings are. We convert the string value to hex, then measure it's length after.
    Then we combine those in reverse - in the style 'string length hex | string hex' and return that so that when
    reading the keytab forwards, the length of the string precedes it.
    The size used to encode the length of the string must also be specified, as different string fields encode their
    length using differently sized integers. Why? Who knows!
    """
    # 1 character = 1 byte so this is easy
    str_byte_length = len(any_string)
    str_length_hex = _write_number_to_hex(str_byte_length, size_of_length_field_bytes, keytab_format_version)
    return str_length_hex + _write_string_to_hex(any_string)


def _write_realm_to_hex(realm: str, keytab_format_version: int):
    """ Write our realm to hex. """
    return _write_string_and_its_length_to_hex(realm, REALM_LENGTH_FIELD_SIZE_BYTES, keytab_format_version)


def _write_principal_component_to_hex(component: str, keytab_format_version: int):
    """ Write a principal component to hex. """
    return _write_string_and_its_length_to_hex(component, PRINCIPAL_COMPONENT_LENGTH_FIELD_SIZE_BYTES, keytab_format_version)


def _write_principal_to_hex(principal: str, keytab_format_version: int):
    """ Write a kerberos principal to hex.
    This means splitting up the components of the principal, the converting the components to hex, measuring each
    component's length in bytes and encoding that length, then concatenating them all and returning.
    """
    spn_components = principal.split(PRINCIPAL_COMPONENT_DIVIDER)
    final_hex = ''
    for component in spn_components:
        final_hex += _write_principal_component_to_hex(component, keytab_format_version)
    return final_hex


def _write_raw_kerberos_key_to_keytab_entry(raw_key, keytab_format_version: int):
    """ Prep our raw kerberos key for inclusion in a keytab entry.
    We can extract the hex from a RawKerberosKey object natively, so we just need to measure its length and then
    add that in front before returning so it can be read back like any other string.
    """
    hex_encoded_key = raw_key.get_raw_hex_encoded_key()
    # 1 hex digit = 4 bits, 1 byte = 8 bits, so to measure the byte length of a hex string we halve its length.
    # keys are always even numbers in length, but just in case, round up non-integers
    key_byte_length = len(hex_encoded_key)/2
    if key_byte_length != int(key_byte_length):
        key_byte_length = math.ceil(key_byte_length)
    key_byte_length = int(key_byte_length)  # cast back to int
    key_length_hex = _write_number_to_hex(key_byte_length, KEY_LENGTH_FIELD_SIZE_BYTES, keytab_format_version)
    return key_length_hex + hex_encoded_key



import copy
import socket
import ssl

from ms_active_directory import logging_utils

from ldap3 import (
    BASE,
    Connection,
    SUBTREE,
)
from typing import List

import ms_active_directory.environment.constants as constants
import ms_active_directory.environment.ldap.ldap_format_utils as ldap_utils
import ms_active_directory.environment.ldap.ldap_constants as ldap_constants
import ms_active_directory.environment.security.security_config_utils as security_utils
import ms_active_directory.environment.security.security_config_constants as security_constants

from ms_active_directory.core.ad_computer import ADComputer
from ms_active_directory.core.ad_users_and_groups import ADGroup, ADUser
from ms_active_directory.exceptions import (
    ObjectCreationException,
    DomainSearchException
)

logger = logging_utils.get_logger()


class ADSession:

    def __init__(self, ldap_connection: Connection, domain):
        self.ldap_connection = ldap_connection
        self.domain = domain
        self.domain_dns_name = self.domain.get_domain_dns_name()
        self.domain_search_base = ldap_utils.construct_ldap_base_dn_from_domain(self.domain_dns_name)

    def is_authenticated(self):
        """ Returns if the session is currently authenticated """
        return self.ldap_connection.bound

    def is_encrypted(self):
        """ Returns if the session's connection is encrypted """
        return self.ldap_connection.tls_started or self.ldap_connection.server.ssl

    def is_open(self):
        """ Returns if the session's connection is currently open """
        return not self.ldap_connection.closed

    def is_thread_safe(self):
        """ Returns if the session's connection is thread-safe """
        return self.ldap_connection.strategy.thread_safe

    def get_ldap_connection(self):
        """ Returns the LDAP connection that this session uses for communication.
        This is particularly useful if a user wants to make complex LDAP queries or perform
        operations that are not supported by the ADSession object, and is willing to craft
        them and parse results themselves.
        """
        return self.ldap_connection

    def get_current_server_uri(self):
        """ Returns the URI of the server that this session is currently communicating with """
        return self.ldap_connection.server.name

    def get_domain(self):
        """ Returns the domain that this session is connected to """
        return self.domain

    def get_domain_dns_name(self):
        """ Returns the domain that this session is connected to """
        return self.domain_dns_name

    def dn_exists_in_domain(self, distinguished_name: str):
        """ Check if a distinguished name exists within the domain, regardless of what it is.
        :param distinguished_name: Either a relative distinguished name or full distinguished name
                                   to search for within the domain.
        :returns: True if the distinguished name exists within the domain.
        """
        # since the distinguished name may be a relative distinguished name or complete one,
        # normalize it
        normalized_rdn = ldap_utils.normalize_object_location_in_domain(distinguished_name,
                                                                        self.domain_dns_name)
        search_dn = normalized_rdn + ',' + self.domain_search_base
        # search returns True if it finds anything. we only need to find one object before stopping
        res = self.ldap_connection.search(search_base=search_dn,
                                          search_filter=ldap_constants.FIND_ANYTHING_FILTER,
                                          search_scope=BASE,
                                          size_limit=1)
        exists, _, _, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        return exists

    def object_exists_in_domain_with_attribute(self, attr: str, unescaped_value: str):
        """ Check if any objects exist in the domain with a given attribute. Returns True if so, False otherwise.
        :param attr: The LDAP attribute to examine in the search.
        :param unescaped_value: The value of the attribute that we're looking for, in its raw form.
        :returns: True if any objects exist in the domain with the attribute specified equal to the value.
        """
        if ldap_utils.is_dn(unescaped_value):
            value = ldap_utils.escape_dn_for_filter(unescaped_value)
            logger.debug('Escaped value %s of %s as distinguished name to be %s', unescaped_value, attr, value)
        else:
            value = ldap_utils.escape_generic_filter_value(unescaped_value)
            logger.debug('Escaped value %s of %s as generic LDAP value to be %s', unescaped_value, attr, value)
        ldap_filter = '({}={})'.format(attr, value)
        # search returns True if it finds anything. we only need to find one object before stopping
        res = self.ldap_connection.search(search_base=self.domain_search_base,
                                          search_filter=ldap_filter,
                                          search_scope=SUBTREE,
                                          attributes=[attr],
                                          size_limit=1)
        _, _, response, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        real_entities = ldap_utils.remove_ad_search_refs(response)
        return len(real_entities) > 0

    def _create_object(self, object_dn: str, object_classes: List[str], account_attr_dict: dict):
        if self.dn_exists_in_domain(object_dn):
            raise ObjectCreationException('An object already exists within the domain with distinguished name {} - '
                                          'please remove it or change the attributes specified such that a different '
                                          'distinguished name is created.'.format(object_dn))
        res = self.ldap_connection.add(object_dn, object_classes, account_attr_dict)
        # TODO: returning the actual response is probably a good idea so we can do something with it.
        # doing more with it in case of error would be a good idea too
        success, result, _, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        if success:
            return success
        # don't include attributes in the exception because a password could be there and it could get logged.
        raise ObjectCreationException('An exception was encountered creating an object with distinguished name {} '
                                      'and object classes {}. LDAP result: {}'.format(object_dn, object_classes, result))

    def create_computer(self, computer_name: str, computer_location: str=None, computer_password: str=None,
                        encryption_types: List[security_constants.ADEncryptionType]=None, hostnames: List[str]=None,
                        services: List[str]=None, supports_legacy_behavior: bool=False,
                        **additional_account_attributes):
        """ Use the session to create a computer in the domain and return a computer object.
        :param computer_name: The common name of the computer to create in the AD domain. This
                              will be used to determine the sAMAccountName, and if no hostnames
                              are specified then this will be used to determine the hostnames for
                              the computer.
        :param computer_location: The distinguished name of the location within the domain where
                                  the computer will be created. It may be a relative distinguished
                                  name (not including the domain component) or a full distinguished
                                  name.  If not specified, defaults to CN=Computers which is
                                  standard for Active Directory.
        :param computer_password: The password to be set for the computer. This is particularly
                                  useful to specify if the computer will be shared across multiple
                                  applications or devices, or if pre-creating a computer for another
                                  application to use. If not specified, a random 120 character
                                  password will be generated.
        :param encryption_types: The encryption types to set as supported on the computer in AD.
                                 These will also be used to generate kerberos keys for the computer.
                                 If not specified, defaults to [aes256-cts-hmac-sha1-96].
        :param hostnames: The hostnames to use for configuring the service principal names of the
                          computer. These may be short hostnames or fully qualified domain names.
                          If not specified, defaults to the "computer_name" as a short hostname and
                          "computer_name.domain" as a fully qualified domain name.
        :param services: The services to enable on each hostname, which will be used with hostnames
                         to generate the computer's service principal names. If not specified,
                         defaults to ["HOST"] which is standard for Active Directory.
        :param supports_legacy_behavior: Does the computer being created support legacy behavior such
                                         as NTLM authentication or UNC path addressing from older windows
                                         clients? Defaults to False. Impacts the restrictions on
                                         computer naming.
        :param additional_account_attributes: Additional LDAP attributes to set on the account and their
                                              values. This is used to support power users setting arbitrary
                                              attributes, such as "userCertificate" to set the certificate
                                              for a computer that will use mutual TLS for EXTERNAL SASL auth.
                                              This also allows overriding of some values that are not explicit
                                              keyword arguments in order to avoid overcomplication, since most
                                              people won't set them (e.g. userAccountControl).
        """
        logger.debug('Request to create computer in domain %s with the following attributes: computer_name=%s, '
                     'computer_location=%s encryption_types=%s hostnames=%s services=%s supports_legacy_behavior=%s '
                     'number of additional attributes specified: %s', self.domain_dns_name, computer_name,
                     computer_location, encryption_types, hostnames, services, supports_legacy_behavior,
                     len(additional_account_attributes))
        # validate our computer name and then determine our sAMAccountName
        computer_name = ldap_utils.validate_and_normalize_computer_name(computer_name, supports_legacy_behavior)
        samaccount_name = computer_name + '$'

        if self.object_exists_in_domain_with_attribute(ldap_constants.AD_ATTRIBUTE_SAMACCOUNT_NAME, samaccount_name):
            raise ObjectCreationException('An object already exists with sAMAccountName {} so a computer may not be '
                                          'created with the name {}'.format(samaccount_name, computer_name))

        # get or normalize our computer location. the end format is as a relative distinguished name
        if computer_location is None:
            computer_location = ldap_constants.DEFAULT_COMPUTER_LOCATION
        else:
            computer_location = ldap_utils.normalize_object_location_in_domain(computer_location,
                                                                               self.domain_dns_name)
        if not self.dn_exists_in_domain(computer_location):
            raise ObjectCreationException('The computer location {} cannot be found in the domain.'
                                          .format(computer_location))
        # now we can build our full object distinguished name
        computer_dn = ldap_utils.construct_object_distinguished_name(computer_name, computer_location,
                                                                     self.domain_dns_name)
        if self.dn_exists_in_domain(computer_dn):
            raise ObjectCreationException('There exists an object in the domain with distinguished name {} and so a '
                                          'computer may not be created in the domain with name {} in location {}. '
                                          'Please use a different name or location.'
                                          .format(computer_dn, computer_name, computer_location))

        # generate a password if needed and encode it
        if computer_password is None:
            computer_password = security_utils.generate_random_ad_password()
        encoded_pw = security_utils.encode_password(computer_password)

        # normalize encryption type values and convert it to the encoded bitstring
        if encryption_types is None:
            encryption_types = [security_constants.ADEncryptionType.AES256_CTS_HMAC_SHA1_96]
        else:
            encryption_types = security_utils.normalize_encryption_type_list(encryption_types)
        encoded_enc_type_value = security_utils.get_supported_encryption_types_value(encryption_types)

        if hostnames is None:
            hostnames = ldap_utils.construct_default_hostnames_for_computer(computer_name, self.domain_dns_name)
        if services is None:
            services = ldap_constants.DEFAULT_COMPUTER_SERVICES
        spns = ldap_utils.construct_service_principal_names(services, hostnames)
        for spn in spns:
            if self.object_exists_in_domain_with_attribute(ldap_constants.AD_ATTRIBUTE_SERVICE_PRINCIPAL_NAMES, spn):
                raise ObjectCreationException('An object exists in the domain with service principal name {} and so '
                                              'creating a computer with the hostnames ({}) and services ({}) in use '
                                              'will cause undefined, conflicting behavior during lookups. Please '
                                              'specify different hostnames or services, or a different computer name '
                                              'if hostnames are not being explicitly set.'
                                              .format(spn, ', '.join(hostnames), ', '.join(services)))

        computer_attributes = {
            ldap_constants.AD_ATTRIBUTE_USER_ACCOUNT_CONTROL: ldap_constants.COMPUTER_ACCESS_CONTROL_VAL,
            ldap_constants.AD_ATTRIBUTE_PASSWORD: encoded_pw,
            ldap_constants.AD_ATTRIBUTE_SAMACCOUNT_NAME: samaccount_name,
            ldap_constants.AD_ATTRIBUTE_ENCRYPTION_TYPES: encoded_enc_type_value,
            ldap_constants.AD_ATTRIBUTE_SERVICE_PRINCIPAL_NAMES: spns,
            ldap_constants.AD_ATTRIBUTE_DNS_HOST_NAME: hostnames[0],
        }
        if len(hostnames) > 1:
            computer_attributes[ldap_constants.AD_ATTRIBUTE_ADDITIONAL_DNS_HOST_NAME] = hostnames[1:]
        # don't include additional attributes for logging in case they're sensitive
        loggable_attributes = copy.deepcopy(computer_attributes)
        del loggable_attributes[ldap_constants.AD_ATTRIBUTE_PASSWORD]
        logger.info('Attempting to create computer in domain %s with the following LDAP attributes: %s and %s additional attributes',
                    loggable_attributes, len(additional_account_attributes))

        # add in our additional account attributes at the end so they can override anything we set here
        computer_attributes.update(additional_account_attributes)

        self._create_object(computer_dn, ldap_constants.OBJECT_CLASSES_FOR_COMPUTER, computer_attributes)
        return ADComputer(samaccount_name, self.domain, computer_location, computer_password, spns,
                          encryption_types)

    def take_over_existing_computer(self, computer_name: str):
        """ Use the session to take over a computer in the domain and return a computer object.
        This resets the computer's password so that nobody else can impersonate it, and reads
        the computer's attributes in order to create a computer object and return it.
        :param computer_name: The common name or sAMAccountName of the computer to find in the AD domain.
                              If it appears to be a common name, not ending in $, a sAMAccountName will
                              be derived to search for. If that cannot be found, then a search will be
                              done for this as a common name. If no unique computer can be found with that
                              search, then an exception will be raised.
        """
        raise NotImplementedError()

    def is_domain_close_in_time_to_localhost(self, allowed_drift_seconds=None):
        """ Get whether the domain time is close to the current local time.
        Just calls the parent domain function and returns that. This is included here for completeness.
        """
        return self.domain.is_close_in_time_to_localhost(self.ldap_connection, allowed_drift_seconds)

    def find_certificate_authorities_for_domain(self, pem_format: bool=True):
        """ Attempt to discover the CAs within the domain and return info on their certificates.
        If a session was first established using an IP address or blind trust TLS, but we want to bootstrap our
        sessions to establish stronger trust, or write the CA certificates to a local truststore for other
        non-LDAP applications to use (e.g. establishing roots of trust for https or syslog over TLS), then it's
        helpful to grab the certificate authorities in the domain and their signing certificates.
        Not all domains run certificate authorities; some use public CAs or get certs from other PKI being run,
        so this isn't useful for everyone. But a lot of people do run CAs in their AD domains, and this is useful
        for them.

        :param pem_format: If True, return the certificates as strings in PEM format. Otherwise, return the
                           certificates as bytestrings in DER format. Defaults to True.
        :returns: A list of either PEM-formatted certificate strings or DER-formatted certificate byte strings,
                  representing the CA certificates of the CAs within the domain.
        """
        ca_filter = '({}={})'.format(ldap_constants.AD_ATTRIBUTE_OBJECT_CLASS,
                                     ldap_constants.CERTIFICATE_AUTHORITY_OBJECT_CLASS)
        search_loc = '{},{}'.format(ldap_constants.DOMAIN_WIDE_CONFIGURATIONS_CONTAINER,
                                    self.domain_search_base)
        res = self.ldap_connection.search(search_base=search_loc, search_filter=ca_filter, search_scope=SUBTREE,
                                          attributes=[ldap_constants.AD_ATTRIBUTE_CA_CERT])
        success, _, entities, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        if not success:
            raise DomainSearchException('Failed to search domain for Certificate Authorities')
        entities = ldap_utils.remove_ad_search_refs(entities)
        logger.info('Found %s CAs within the domain', len(entities))
        ca_certs_der_fmt = []
        for entity in entities:
            # the bytes are returned in a 1-item list
            ca_cert_der_bytes = entity['attributes'][ldap_constants.AD_ATTRIBUTE_CA_CERT][0]
            ca_certs_der_fmt.append(ca_cert_der_bytes)

        if pem_format:
            return [ssl.DER_cert_to_PEM_cert(cert_bytes) for cert_bytes in ca_certs_der_fmt]
        return ca_certs_der_fmt

    def find_current_time_for_domain(self):
        """ Get the current time for the domain as a datetime object.
        Just calls the parent domain function and returns that. This is included here for completeness.
        :returns: A datetime object representing the current time in the domain.
        """
        return self.domain.find_current_time(self.ldap_connection)

    def find_dns_servers_for_domain(self):
        """ Attempt to discover the DNS servers within the domain and return info on them.
        If a session was first established using an IP address or blind trust TLS, but we want to bootstrap our
        sessions to use kerberos or TLS backed by CA certificates, we need proper DNS configured. For private
        domains (e.g. in a datacenter), we may run DNS servers within the domain. This function discovers
        computers with a "DNS/" service principal name, tries to look up IP addresses for them, and then
        returns that information.
        This won't always be useful, as DNS isn't always part of the AD domain, but it can help if we're bootstrapping
        a computer with manufacturer configurations to use the AD domain for everything based on a minimal starting
        configuration.

        :returns: A dictionary mapping DNS hostnames of DNS servers to IP addresses. The hostnames are provided in case
                  a caller is configuring DNS-over-TLS. If no IP address can be resolved for a hostname, it will map to
                  a None value.
                  https://datatracker.ietf.org/doc/html/rfc8310
        """
        dns_computer_filter = '(&({}={})({}={}))'.format(ldap_constants.AD_ATTRIBUTE_OBJECT_CLASS,
                                                         ldap_constants.COMPUTER_OBJECT_CLASS,
                                                         ldap_constants.AD_ATTRIBUTE_SERVICE_PRINCIPAL_NAMES,
                                                         ldap_constants.DNS_SERVICE_FILTER)
        # search the whole domain and grab the dns hostnames of the computers found
        res = self.ldap_connection.search(search_base=self.domain_search_base, search_filter=dns_computer_filter,
                                          search_scope=SUBTREE, attributes=[ldap_constants.AD_ATTRIBUTE_DNS_HOST_NAME])
        success, _, entities, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        if not success:
            raise DomainSearchException('Failed to search domain for Computers hosting DNS services')
        entities = ldap_utils.remove_ad_search_refs(entities)
        logger.info('Found %s computers that host DNS services within the domain', len(entities))

        dns_results = {}
        for entity in entities:
            hostname = entity['attributes'][ldap_constants.AD_ATTRIBUTE_DNS_HOST_NAME]
            ip_addr = None
            try:
                ip_addr = socket.gethostbyname(hostname)
            except:
                logger.warning('Unable to resolve ip address for %s', hostname)

            if hostname in dns_results:
                logger.warning('Multiple computer entities exist in the domain serving DNS services with the DNS '
                               'hostname %s - this may indicate a domain misconfiguration issue.', hostname)
            dns_results[hostname] = ip_addr
        return dns_results

    def find_forest_schema_version(self):
        """ Attempt to determine the version of Windows Server set in the forest's schema.
        returns: An Enum of type ADVersion indicating the schema version.
        """
        search_loc = '{},{}'.format(ldap_constants.DOMAIN_CONTROLLER_SCHEMA_VERSION_SEARCH_CONTAINER,
                                    self.domain_search_base)
        res = self.ldap_connection.search(search_base=search_loc, search_filter=ldap_constants.FIND_ANYTHING_FILTER,
                                          search_scope=BASE, attributes=[ldap_constants.AD_ATTRIBUTE_GET_ALL_NON_VIRTUAL_ATTRS])
        success, _, entities, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        if not success:
            raise DomainSearchException('Failed to search domain for schema.')
        entities = ldap_utils.remove_ad_search_refs(entities)
        if len(entities) == 0:
            raise DomainSearchException('The forest schema could not be found when searching the domain.')
        schema = entities[0]
        ad_schema_ver = schema['attributes'][ldap_constants.AD_SCHEMA_VERSION]
        return constants.ADVersion.get_version_from_schema_number(ad_schema_ver)

    def find_functional_level_for_domain(self):
        """ Attempt to discover the functional level of the domain and return it.
        This will indicate if the domain is operating at the level of a 2008, 2012R2, 2016, etc. domain.
        The functional level of a domain influences what functionality exists (e.g. 2003 cannot issue AES keys,
        2012 cannot use many TLS ciphers introduced with TLS1.3) and so it can be useful for determining what
        to do.
        :returns: An Enum of type ADFunctionalLevel indicating the functional level.
        """
        return self.domain.find_functional_level(self.ldap_connection)

    def find_supported_sasl_mechanisms_for_domain(self):
        """ Attempt to discover the SASL mechanisms supported by the domain and return them.
        This just builds upon the functionality that the domain has for this, as you don't need
        to be authenticated as anything other than anonymous to read this information (since it's
        often used to figure out how to authenticate).
        This is included in the session object for completeness.
        """
        return self.domain.find_supported_sasl_mechanisms(self.ldap_connection)

    def _find_user_or_group_and_attrs(self, search_base: str, search_filter:str, search_scope: str,
                                      attributes: List[str], size_limit: int, return_type):
        """ A helper function for common search and result parsing logic in other find functions for users
        and groups
        """
        attrs = self._figure_out_search_attributes_for_user_or_group(attributes)
        res = self.ldap_connection.search(search_base=search_base,
                                          search_filter=search_filter,
                                          search_scope=search_scope,
                                          size_limit=size_limit,
                                          attributes=attrs)
        _, _, resp, _ = ldap_utils.process_ldap3_conn_return_value(self.ldap_connection, res)
        resp = ldap_utils.remove_ad_search_refs(resp)
        if not resp:
            return []

        results = []
        for entry in resp:
            entry_attributes = entry['attributes']
            obj = return_type(entry['dn'], entry_attributes, self.domain)
            results.append(obj)
        return results

    def find_groups_by_common_name(self, group_name: str, attributes_to_lookup: List[str]=None):
        """ Find all groups with a given common name and return a list of ADGroup objects.
        This is particularly useful when you have multiple groups with the same name in different OUs
        as a result of a migration, and want to find them so you can combine them.

        :param group_name: The common name of the group(s) to be looked up.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the group. Regardless of
                                     what's specified, the groups' name and object class attributes will be queried.
        :returns: a list of ADGroup objects representing groups with the specified common name.
        """
        # build a compound filter for users with this common name
        search_filter = '(&({cn_attr}={cn}){type_filter})'.format(cn_attr=ldap_constants.AD_ATTRIBUTE_COMMON_NAME,
                                                                  cn=group_name,
                                                                  type_filter=ldap_constants.FIND_GROUP_FILTER)
        # a size limit of 0 means unlimited
        res = self._find_user_or_group_and_attrs(self.domain_search_base, search_filter, SUBTREE,
                                                 attributes_to_lookup, 0, ADGroup)
        logger.info('%s groups found with common name %s', len(res), group_name)
        return res

    def find_group_by_distinguished_name(self, group_dn: str, attributes_to_lookup: List[str]=None):
        """ Find a group in AD based on a specified distinguished name and return it along with any
        requested attributes.
        :param group_dn: The distinguished name of the group.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the group. Regardless of
                                     what's specified, the group's name and object class attributes will be queried.
        :returns: an ADGroup object or None if the group does not exist.
        """
        # since the distinguished name may be a relative distinguished name or complete one,
        # normalize it
        normalized_rdn = ldap_utils.normalize_object_location_in_domain(group_dn,
                                                                        self.domain_dns_name)
        search_dn = normalized_rdn + ',' + self.domain_search_base
        res = self._find_user_or_group_and_attrs(search_dn, ldap_constants.FIND_GROUP_FILTER, BASE,
                                                 attributes_to_lookup, 1, ADGroup)
        if not res:
            return None
        return res[0]

    def find_group_by_sam_name(self, group_name, attributes_to_lookup=None):
        """ Find a Group in AD based on a specified sAMAccountName name and return it along with any
        requested attributes.
        :param group_name: The sAMAccountName name of the group.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the group. Regardless of
                                     what's specified, the user's name and object class attributes will be queried.
        :returns: an ADGroup object or None if the user does not exist.
        """
        # build a compound filter for users with this samaccountname
        search_filter = '(&({sam_attr}={sam_name}){type_filter})'.format(sam_attr=ldap_constants.AD_ATTRIBUTE_SAMACCOUNT_NAME,
                                                                         sam_name=group_name,
                                                                         type_filter=ldap_constants.FIND_GROUP_FILTER)
        res = self._find_user_or_group_and_attrs(self.domain_search_base, search_filter, SUBTREE,
                                                 attributes_to_lookup, 1, ADGroup)
        if not res:
            return None
        return res[0]

    def find_users_by_common_name(self, user_name: str, attributes_to_lookup: List[str]=None):
        """ Find all users with a given common name and return a list of ADUser objects.
        This is particularly useful when you have multiple users with the same name in different OUs
        as a result of a migration, and want to find them so you can combine them.

        :param user_name: The common name of the user(s) to be looked up.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the users. Regardless of
                                     what's specified, the users' name and object class attributes will be queried.
        :returns: a list of ADUser objects representing users with the specified common name.
        """
        # build a compound filter for users with this common name
        search_filter = '(&({cn_attr}={cn}){type_filter})'.format(cn_attr=ldap_constants.AD_ATTRIBUTE_COMMON_NAME,
                                                                  cn=user_name,
                                                                  type_filter=ldap_constants.FIND_USER_FILTER)
        # a size limit of 0 means unlimited
        res = self._find_user_or_group_and_attrs(self.domain_search_base, search_filter, SUBTREE,
                                                 attributes_to_lookup, 0, ADUser)
        logger.info('%s users found with common name %s', len(res), user_name)
        return res

    def find_user_by_distinguished_name(self, user_dn: str, attributes_to_lookup: List[str]=None):
        """ Find a User in AD based on a specified distinguished name and return it along with any
        requested attributes.
        :param user_dn: The distinguished name of the user.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the user. Regardless of
                                     what's specified, the user's name and object class attributes will be queried.
        :returns: an ADUser object or None if the user does not exist.
        """
        # since the distinguished name may be a relative distinguished name or complete one,
        # normalize it
        normalized_rdn = ldap_utils.normalize_object_location_in_domain(user_dn,
                                                                        self.domain_dns_name)
        search_dn = normalized_rdn + ',' + self.domain_search_base
        res = self._find_user_or_group_and_attrs(search_dn, ldap_constants.FIND_USER_FILTER, BASE,
                                                 attributes_to_lookup, 1, ADUser)
        if not res:
            return None
        return res[0]

    def find_user_by_sam_name(self, user_name, attributes_to_lookup: List[str]=None):
        """ Find a User in AD based on a specified sAMAccountName name and return it along with any
        requested attributes.
        :param user_name: The sAMAccountName name of the user.
        :param attributes_to_lookup: A list of additional LDAP attributes to query for the user. Regardless of
                                     what's specified, the user's name and object class attributes will be queried.
        :returns: an ADUser object or None if the user does not exist.
        """
        # build a compound filter for users with this samaccountname
        search_filter = '(&({sam_attr}={sam_name}){type_filter})'.format(sam_attr=ldap_constants.AD_ATTRIBUTE_SAMACCOUNT_NAME,
                                                                         sam_name=user_name,
                                                                         type_filter=ldap_constants.FIND_USER_FILTER)
        res = self._find_user_or_group_and_attrs(self.domain_search_base, search_filter, SUBTREE,
                                                 attributes_to_lookup, 1, ADUser)
        if not res:
            return None
        return res[0]

    def _figure_out_search_attributes_for_user_or_group(self, attributes_to_lookup):
        """ There's some attributes we'll always get for users and groups, whether callers requested them or not.
        This combines those with any requested attributes
        """
        base_group_attrs = {ldap_constants.AD_ATTRIBUTE_SAMACCOUNT_NAME, ldap_constants.AD_ATTRIBUTE_OBJECT_CLASS,
                            ldap_constants.AD_ATTRIBUTE_COMMON_NAME}
        if attributes_to_lookup:
            base_group_attrs.update(set(base_group_attrs))
        # sort for reproducibility in eventual testing
        return sorted(list(base_group_attrs))



from ms_active_directory import logging_utils

from ms_active_directory.environment.kerberos.kerberos_constants import (
    AD_ENC_TYPE_TO_KRB5_ENC_TYPE_MAP,
    DEFAULT_UNKNOWN_NAME_TYPE,
    KRB5_ENC_TYPE_TO_AD_ENC_TYPE_MAP,
    NAME_TYPE_VALUE_TO_NAME_TYPE_MAP,
    PREFERRED_KEYTAB_FORMAT_VERSION
)
from ms_active_directory.environment.kerberos.kerberos_keytab_generator import write_complete_keytab_structure_to_raw_bytes
from ms_active_directory.environment.security.security_config_constants import (
    ADEncryptionType
)


logger = logging_utils.get_logger()


class RawKerberosKey:
    """ A raw kerberos key - containing only the generated shared secret and the encryption type.
    This does not contain any information about who's using it, its purpose, etc. and is tied
    only to the password used, the salt, and the encryption type. It can therefore be used to
    generate usable kerberos keys for either accepting or initiating GSS authentication.
    """
    def __init__(self, enc_type, key_bytes: bytes):
        # leave enc_type flexible. it can be either an ADEncryption type or a string depending on whether
        # we generate this key ourselves for an AD account or read it in from a keytab file
        self.encryption_type = enc_type
        self.key_bytes = key_bytes
        # hex is the most common key representation for visual display, so we have it by default
        self.raw_hex_encoded_key = key_bytes.hex()
        self.hex_encoded_key = '0x' + self.raw_hex_encoded_key

    def get_key_bytes(self):
        return self.key_bytes

    def get_hex_encoded_key(self):
        return self.hex_encoded_key

    def get_raw_hex_encoded_key(self):
        return self.raw_hex_encoded_key

    def uses_active_directory_supported_encryption_type(self):
        return isinstance(self.encryption_type, ADEncryptionType) or self.encryption_type in KRB5_ENC_TYPE_TO_AD_ENC_TYPE_MAP


class GssKerberosKey:
    """ A kerberos key that can actually be used in kerberos negotiation (as either a user or a server).
    This is a raw key properly wrapped with encoded additional information about the principal, kvno,
    encryption type, etc.
    """

    def __init__(self, principal: str, realm: str, raw_key: RawKerberosKey, kvno: int, flags: int=None,
                 timestamp: int=None, gss_name_type: int=DEFAULT_UNKNOWN_NAME_TYPE,
                 format_version: int=PREFERRED_KEYTAB_FORMAT_VERSION):
        self.principal = principal
        self.realm = realm
        self.raw_key = raw_key
        # technically, for AD-only purposes, we could have typed this as an ADEncryptionType.
        # but since this is built directly from a keytab file in some cases, we might have
        # read in something else
        enc_type = self.raw_key.encryption_type
        if isinstance(enc_type, ADEncryptionType):
            enc_type = AD_ENC_TYPE_TO_KRB5_ENC_TYPE_MAP[enc_type]
        self.encryption_type = enc_type
        self.kvno = kvno
        self.flags = flags
        self.timestamp = timestamp
        self.gss_name_type = gss_name_type
        self.friendly_gss_name_type = None
        # allow for unrecognized gss name type values just in case something new comes out.
        # this field is flavor text anyway and doesn't require functional support, so any
        # value that actually works with the kerberos protocol is fine to use even if we
        # don't have a friendly name for it
        if self.gss_name_type in NAME_TYPE_VALUE_TO_NAME_TYPE_MAP:
            self.friendly_gss_name_type = NAME_TYPE_VALUE_TO_NAME_TYPE_MAP[gss_name_type]
        self.format_version = format_version
        self.complete_keytab_bytes = None

    def get_complete_keytab_bytes(self, format_version: int=None, use_current_time: bool=None):
        """ Get this key object encoded as the bytes of a complete, usable keytab that can be written
        to a file and used for kerberos authentication (initiating or accepting contexts).
        :param format_version: An keytab format version. If not specified, defaults to the format version
                               in the object. If the object's format version is null, defaults to 2.
        :param use_current_time: Whether or not the current time should be used as the timestamp in the
                                 keytab produced, overwriting the time in the object. If no timestamp is
                                 in the object, the current time is used. Defaults to False if not specified.
        """
        return write_complete_keytab_structure_to_raw_bytes(self, keytab_format_version=format_version,
                                                            use_current_time_for_keytab=use_current_time)

    def get_raw_key_bytes(self):
        return self.raw_key.get_key_bytes()

    def uses_active_directory_supported_encryption_type(self):
        return self.encryption_type in KRB5_ENC_TYPE_TO_AD_ENC_TYPE_MAP

    def set_principal(self, principal: str):
        """ Sets the principal and clears complete_gss_keytab_bytes so we re-compute it """
        self.principal = principal
        self.complete_keytab_bytes = None

    def set_realm(self, realm: str):
        """ Sets the realm and clears complete_gss_keytab_bytes so we re-compute it """
        self.realm = realm
        self.complete_keytab_bytes = None

    def set_raw_key(self, raw_key: RawKerberosKey):
        """ Sets the raw key, updates our encryption type and clears complete_gss_keytab_bytes so we re-compute it.
        The encryption type is directly tied to our raw key and vice versa, so setting one without the other makes no
        sense.
        """
        self.raw_key = raw_key
        enc_type = self.raw_key.encryption_type
        if isinstance(enc_type, ADEncryptionType):
            enc_type = AD_ENC_TYPE_TO_KRB5_ENC_TYPE_MAP[enc_type]
        self.encryption_type = enc_type
        self.complete_keytab_bytes = None

    def set_kvno(self, kvno: int):
        """ Sets the kvno and clears complete_gss_keytab_bytes so we re-compute it """
        self.kvno = kvno
        self.complete_keytab_bytes = None

    def set_flags(self, flags: int):
        """ Sets the flags and clears complete_gss_keytab_bytes so we re-compute it """
        self.flags = flags
        self.complete_keytab_bytes = None

    def set_timestamp(self, timestamp: int):
        """ Sets the timestamp and clears complete_gss_keytab_bytes so we re-compute it """
        self.timestamp = timestamp
        self.complete_keytab_bytes = None

    def set_gss_name_type(self, name_type: int):
        """ Sets the gss name type and friendly name type and clears complete_gss_keytab_bytes so we re-compute it """
        self.gss_name_type = name_type
        self.friendly_gss_name_type = NAME_TYPE_VALUE_TO_NAME_TYPE_MAP[name_type]
        self.complete_keytab_bytes = None

    def set_format_version(self, format_version: int):
        """ Sets the keytab format version and clears complete_gss_keytab_bytes so we re-compute it """
        self.format_version = format_version
        self.complete_keytab_bytes = None



""" Kerberos Key Generator Module

This module contains functions to generate kerberos keys based on a given encryption
type and password bytes.
Optionally, the bytes for a salt may be supplied, and iteration count may be specified
for those encryption types that support salting or multiple iterations.
"""
from ms_active_directory import logging_utils

from functools import reduce

from Crypto.Util.number import GCD
from Crypto.Cipher import AES
from Crypto.Hash import HMAC, MD4, SHA  # SHA is SHA1, not SHA256 or SHA384
from Crypto.Protocol.KDF import PBKDF2  # password-based key derivation function

from ms_active_directory.core.ad_kerberos_keys import RawKerberosKey

from ms_active_directory.environment.kerberos.kerberos_constants import (
    AD_ENC_TYPE_TO_KRB5_ENC_TYPE_MAP,
    AES_CIPHER_BLOCK_SIZE_BYTES,
    AES_ITERATIONS_FOR_AD,
    SALT_FORMAT_FOR_AD,
)
from ms_active_directory.environment.security.security_config_constants import (
    ADEncryptionType,
    ENCRYPTION_TYPE_STR_TO_ENUM,
)

logger = logging_utils.get_logger()


def ad_password_string_to_key(ad_encryption_type: ADEncryptionType, ad_computer_name:str, ad_password: str,
                              ad_domain_dns_name: str, ad_auth_realm: str=None):
    """ Given an encryption type, a computer name, a password, and a domain, generate the raw kerberos key for an AD
    account. Optionally, a realm may be specified if the kerberos realm for the domain is not the domain itself
    (this may be the case for subdomains or when AD is not the central authentication for an environment).
    :param ad_encryption_type: The kerberos encryption type to use for generating the key.
    :param ad_computer_name: The name of the computer in AD. This is the sAMAccountName without the trailing $.
    :param ad_password: The password of the computer.
    :param ad_domain_dns_name: The DNS name of the AD domain where the computer exists.
    :param ad_auth_realm: The realm used by the domain for authentication. If not specified, defaults to the domain
                          in all captial letters.
    """
    ad_auth_realm = ad_auth_realm if ad_auth_realm else ad_domain_dns_name
    # be forgiving to those who don't read the docs
    if ad_computer_name.endswith('$'):
        ad_computer_name = ad_computer_name[:-1]
    salt_str = _format_aes_salt_for_ad(ad_computer_name, ad_domain_dns_name, ad_auth_realm)
    # we can just always pass in a salt and iterations, and unsalted encryption types (e.g. rc4-hmac) will ignore it
    return password_string_to_key(ad_encryption_type, ad_password, salt_str, AES_ITERATIONS_FOR_AD)


def password_bytes_to_key(ad_encryption_type: ADEncryptionType, password_bytes: bytes, salt_bytes: bytes=None,
                          iterations: int=None):
    """ Given an encryption type, password bytes, and optionally salt bytes and an iteration count, generate and
    return a kerberos key for the specified encryption type using the other parameters.
    """
    e = _get_enc_type_profile(ad_encryption_type)
    return e.password_bytes_to_key(password_bytes, salt_bytes, iterations)


def password_string_to_key(ad_encryption_type: ADEncryptionType, password_string: str, salt_string: str=None,
                           iterations: int=None):
    """ Given an encryption type, a string password, and optionally a string salt and an iteration count, generate and
    return a kerberos key for the specified encryption type using the other parameters.
    """
    password_bytes = password_string.encode('UTF-8')
    salt_bytes = salt_string.encode('UTF-8') if salt_string is not None else salt_string
    return password_bytes_to_key(ad_encryption_type, password_bytes, salt_bytes=salt_bytes, iterations=iterations)


def _get_enc_type_profile(enc_type: ADEncryptionType):
    if isinstance(enc_type, str):
        enc_type = ENCRYPTION_TYPE_STR_TO_ENUM.get(enc_type.lower())
    if enc_type not in AD_ENC_TYPE_TO_KRB5_ENC_TYPE_MAP:
        raise ValueError('Invalid or unsupported encryption type for kerberos key generation {}'
                         .format(enc_type))
    ad_enc_to_profile = {
        _AES128CTS.enc_type: _AES128CTS,
        _AES256CTS.enc_type: _AES256CTS,
        _RC4.enc_type: _RC4
    }
    return ad_enc_to_profile[enc_type]


def _format_aes_salt_for_ad(computer_name: str, domain: str, realm: str):
    """ Computer names and domains can be specified in any casing. However, DNS names are case insensitive, as are
    computer names in AD.
    However, salts for AES encryption are not case insensitive because AES doesn't cater to Active Directory's desires.
    The result is a confusing standard that we try to gracefully accommodate by converting casing as needed so that we,
    as a client, can generate keys the same way AD does internally.
    AD also uses different information in its salt than other kerberos realms, so we handle that here as well.
    Normal kerberos realms just do [case sensitive realm][case sensitive principal], which runs contrary to AD's notion
    of domain name/realm name and principals directly translating into DNS and therefore being case insensitive.
    """
    # in the AES salt for AD, the realm piece is always uppercase
    upper_realm = realm.upper()
    # in the AES salt for AD, the domain is always lowercase
    lower_domain = domain.lower()
    # active directory names are case insensitive, so AD decided that salts should always use the lower case name for
    # the computer, because the salting of AES is not case insensitive
    lower_computer_name = computer_name.lower()
    salt_string = SALT_FORMAT_FOR_AD.format(lowercase_computer_name=lower_computer_name,
                                            uppercase_realm=upper_realm,
                                            lowercase_domain=lower_domain)
    return salt_string


def _zeropad(byte_string: bytes, pad_size: int):
    # Return byte_string padded with 0 bytes to a multiple of pad_size.
    padlen = (pad_size - (len(byte_string) % pad_size)) % pad_size
    return byte_string + b'\0'*padlen


def _nfold(string_to_fold: str, nbytes: int):
    """ This function is really hard to read because heavy math doesn't translate super well to
    python. This would actually be more readable in python2 where you can freely float between
    bytes and strings (which might be more readable computing byte-based functions using
    shared secrets that are strings).
    This function applies the n-fold operation defined in RFC3961 to a byte string of length
    nbytes
    https://tools.ietf.org/html/rfc3961#section-5.1
    """
    def _rotate_right(byte_string_to_rotate: str, nbits: int):
        """ Rotate the bytes in str to the right by nbits bits. """
        indices = list(range(len(byte_string_to_rotate)))
        num_bytes, remain = (nbits//8) % len(byte_string_to_rotate), nbits % 8
        return b''.join(bytes([(ord(byte_string_to_rotate[i-num_bytes]) >> remain) |
                               ((ord(byte_string_to_rotate[i-num_bytes-1]) << (8-remain)) & 0xff)])
                        for i in indices)

    def _add_ones_complement(byte_str_1: bytes, byte_str_2: bytes):
        """ Add equal-length strings together with end-around carry. """
        n = len(byte_str_1)
        v = [a + b for a, b in list(zip(byte_str_1, byte_str_2))]
        # Propagate carry bits to the left until there aren't any left.
        while any(x & ~0xff for x in v):
            v = [(v[i-n+1]>>8) + (v[i]&0xff) for i in list(range(n))]
        return b''.join(bytes([x]) for x in v)

    # Concatenate copies of string_to_fold to produce the least common multiple
    # of len(string_to_fold) and nbytes, rotating each copy of string_to_fold to the right
    # by 13 bits times its list position.
    byte_length_to_fold = len(string_to_fold)
    lcm = int(nbytes * byte_length_to_fold / GCD(nbytes, byte_length_to_fold))
    big_byte_str = b''.join((_rotate_right(string_to_fold, 13 * i) for i in range(lcm // byte_length_to_fold)))
    # Decompose the concatenation into slices of length nbytes, and add them together as
    # big-endian ones' complement integers.
    slices = [big_byte_str[p:p+nbytes] for p in list(range(0, lcm, nbytes))]
    result = reduce(_add_ones_complement, slices)
    return result


class _EncTypeProfile(object):
    # Base class for enc_type profiles.  Usable enc_type classes must define:
    #   * enc_type: enc_type number
    #   * key_size: protocol size of key in bytes
    #   * seed_size: random_to_key input size in bytes
    #   * password_bytes_to_key
    enc_type = None
    key_size = None
    seed_size = None

    @classmethod
    def password_bytes_to_key(cls, password_bytes: bytes, salt_bytes: bytes=None, iterations: int=None):
        raise NotImplementedError('Child classes must implement password_bytes_to_key')


class _SimplifiedEnctype(_EncTypeProfile):
    # Base class for enc_types using the RFC 3961 simplified profile.
    # Defines the encrypt, decrypt, and prf methods.  Subclasses must
    # define:
    #   * block_size: The block size used for any CBC computation
    #   * basic_encrypt: Underlying CBC/CTS cipher, which is used for some key generation
    block_size = 1

    @classmethod
    def basic_encrypt(clscls, key: RawKerberosKey, plaintext_bytes: bytes):
        """ Placeholder to force child classes to implement this """
        raise NotImplementedError('This function must be implemented by child classes that need key derivation or encryption')

    @classmethod
    def derive(cls, key: RawKerberosKey, constant: str):
        """ Derive a kerberos key from some key and some constant.
        By mixing a key with a constant, you can essentially make the key
        "service-unique" so that if I use a password to generate a kerberos
        key as well as a key for some other service, they will not be the
        same key, even if they use the same encryption type.
        """
        rnd_seed = b''
        # RFC 3961 says to n-fold the constant only if it is shorter than the cipher block size
        # N-fold is a no-op if the constant is equal to the cipher block size.
        # And unix implementations all n-fold when the constant is greater than the cipher
        # block size. The RFC also gives examples of n-folding both shorter and longer constants
        # So we just always n-fold the constant here
        plaintext_bytes = _nfold(constant, cls.block_size)
        while len(rnd_seed) < cls.seed_size:
            ciphertext_bytes = cls.basic_encrypt(key, plaintext_bytes)
            rnd_seed += ciphertext_bytes
            # use our ciphertext output as the next plaintext input
            plaintext_bytes = ciphertext_bytes
        # trim our generated encrypted seed to form our key_bytes
        trimmed_seed = rnd_seed[:cls.seed_size]
        return RawKerberosKey(cls.enc_type, trimmed_seed)


class _AESEnctype(_SimplifiedEnctype):
    # Base class for aes128-cts and aes256-cts.
    # cts = cipher-text stealing
    # AD only supports aes256-sha1 and aes128-sha1 right now. but if we do generic kdc support, we may add sha256/sha384
    # as options for other AES types, so make this a class variable.
    # AD might also add sha256/sha384 some day. SHA1 still has weak collision resistance, making it suitable for
    # kerberos keys, even though it was broken for strong collision resistance in 2019. but it could be broken for
    # weak collision resistance soon, and so newer AD may very likely start supporting sha256 and sha384
    sha_version = SHA
    # AES candidates, regardless of key size, always uses the same block size, which is 16 bytes
    # this is formalized in RFC3602
    block_size = AES_CIPHER_BLOCK_SIZE_BYTES

    @classmethod
    def password_bytes_to_key(cls, password_bytes: bytes, salt_bytes: bytes, iterations: int=1):
        # this is the pseudo-random function needed for our password-based key
        # derivation for AES
        prf = lambda p, s: HMAC.new(p, s, cls.sha_version).digest()
        # we then compute our seed by running that pseudo-random function
        # over our password bytes and salt bytes for some number of iterations.
        # each subsequent iteration's input will be based on the previous
        # iterations output. this is the basic premise of the AES algorithm.
        # Seed size varies between AES128, AES256, etc. and influences the size
        # of the seed (in bytes) that we pack into.
        # PBKDF2 can take a bytestring or string, so ignore the warning about this
        # being bytes while a string is expected
        seed = PBKDF2(password_bytes, salt_bytes, cls.seed_size, iterations, prf)
        # our first key is a temporary key based off our our seed, which was
        # generated from our password
        tkey = RawKerberosKey(cls.enc_type, seed)
        # our final computed key is a derivation of our temporary key with a usage
        # string 'kerberos', which ensures that AES-based keys for other protocols
        # will be different, even with the same password and salt.
        # this element of AES is a big reason why it's better to save kerberos
        # keys than passwords - because getting the password gives you access over
        # every protocol, while getting a key only gives you kerberos.
        return cls.derive(tkey, 'kerberos')

    @classmethod
    def basic_encrypt(cls, key: RawKerberosKey, plaintext_bytes: bytes):
        # CBC = cipher block chaining
        aes = AES.new(key.key_bytes, AES.MODE_CBC, b'\0' * cls.block_size)
        # pad our plaintext to a multiple of the AES block size
        ctext = aes.encrypt(_zeropad(plaintext_bytes, cls.block_size))
        if len(plaintext_bytes) > cls.block_size:
            # Swap the last two ciphertext blocks and truncate the
            # final block to match the plaintext length.
            lastlen = len(plaintext_bytes) % cls.block_size or cls.block_size
            last_block_index = -1*cls.block_size
            second_to_last_block_index = -2*cls.block_size
            truncated_second_to_last_block = ctext[second_to_last_block_index:second_to_last_block_index][:lastlen]
            ctext = ctext[:second_to_last_block_index] + ctext[last_block_index:] + truncated_second_to_last_block
        return ctext


class _AES128CTS(_AESEnctype):
    enc_type = ADEncryptionType.AES128_CTS_HMAC_SHA1_96
    # 16 bytes * 8 bits/byte = 128 bits, AES128
    key_size = 16
    seed_size = 16


class _AES256CTS(_AESEnctype):
    enc_type = ADEncryptionType.AES256_CTS_HMAC_SHA1_96
    # 32 bytes * 8 bits/byte = 256 bits, AES256
    key_size = 32
    seed_size = 32


class _RC4(_EncTypeProfile):
    enc_type = ADEncryptionType.RC4_HMAC
    key_size = 16
    seed_size = 16

    @classmethod
    def password_bytes_to_key(cls, password_bytes: bytes, salt_bytes: bytes=None, iterations: bytes=None):
        # RC4 requires a shared secret that fits in the UTF8 encoding, which is
        # part of why it's considered a little less secure (smaller space).
        # we then convert the shared secret to utf16 (little-endian) and generate
        # they key by computing the MD4 digest of that utf16 string.
        # This is very straightforward and easy.
        utf16_string = password_bytes.decode('UTF-8').encode('UTF-16LE')
        return RawKerberosKey(cls.enc_type, MD4.new(utf16_string).digest())



from ms_active_directory.core.ad_computer import (
    ADComputer
)

from ms_active_directory.core.ad_domain import (
    ADDomain,
    join_ad_domain,
    join_ad_domain_using_session,
)

from ms_active_directory.core.ad_kerberos_keys import (
    GssKerberosKey
)

from ms_active_directory.core.ad_session import (
    ADSession,
)

from ms_active_directory.core.ad_users_and_groups import (
    ADGroup,
    ADUser
)

from ms_active_directory.environment.kerberos.kerberos_keytab_generator import (
    write_gss_kerberos_key_list_to_raw_bytes,
    write_gss_kerberos_key_list_to_raw_hex,
)
from ms_active_directory.environment.kerberos.kerberos_keytab_ingester import (
    process_keytab_file_to_extract_entries,
)
from ms_active_directory.environment.kerberos.kerberos_raw_key_generator import (
    ad_password_string_to_key,
    password_string_to_key
)

from ms_active_directory.environment.ldap.ldap_constants import *

from ms_active_directory.environment.security.security_config_constants import (
    ADEncryptionType,
)

from ms_active_directory.environment.constants import (
    ADFunctionalLevel,
    ADVersion,
)

from ms_active_directory.exceptions import *
from ms_active_directory.logging_utils import configure_log_level



""" Exceptions used within the library """


class MsActiveDirectoryException(Exception):
    """ A parent class for all other exceptions so that users can have a catch-all exception for
    functional issues that still doesn't blind them to things like accidentally providing a string
    where a number is needed.
    """
    def __init__(self, exception_str):
        self.message = exception_str
        super().__init__(self.message)


class DomainConnectException(MsActiveDirectoryException):
    def __init__(self, exception_str):
        super().__init__(exception_str)


class DomainSearchException(MsActiveDirectoryException):
    def __init__(self, exception_str):
        super().__init__(exception_str)


class InvalidComputerParameterException(MsActiveDirectoryException):
    def __init__(self, exception_str):
        super().__init__(exception_str)


class InvalidDomainParameterException(MsActiveDirectoryException):
    def __init__(self, exception_str):
        super().__init__(exception_str)


class InvalidLdapParameterException(MsActiveDirectoryException):
    def __init__(self, exception_str):
        super().__init__(exception_str)


class KeytabEncodingException(MsActiveDirectoryException):
    def __init__(self, exception_str):
        super().__init__(exception_str)


class LdapResponseDecodeException(MsActiveDirectoryException):
    def __init__(self, exception_str):
        super().__init__(exception_str)


class ObjectCreationException(MsActiveDirectoryException):
    def __init__(self, exception_str):
        super().__init__(exception_str)



""" Constants for discovering and interacting with an AD domain and the environment around it """

LDAP_DNS_SRV_FORMAT = '_ldap._tcp.dc._msdcs.{domain}'
KERBEROS_DNS_SRV_FORMAT = '_kerberos._tcp.dc._msdcs.{domain}'

LDAP_SITE_AWARE_DNS_SRV_FORMAT = '_ldap._tcp.{site}._sites.dc._msdcs.{domain}'
KERBEROS_SITE_AWARE_DNS_SRV_FORMAT = '_ldap._tcp.{site}._sites.dc._msdcs.{domain}'

DNS_TIMEOUT_SECONDS = 10



""" This module contains helper functions for discovering an Active Directory domain and its
domain controllers via DNS, as well as for sorting them by reachability.
"""
import dns.exception
import dns.resolver
import socket
import time

import ms_active_directory.environment.format_utils as format_utils

# environmental interactions are lightweight and primarily IO-bounded, not CPU-bounded.
# most of our time is spent waiting on replies, so we use a thread pool instead of a
# process pool
from concurrent.futures import ThreadPoolExecutor
from dns.rdatatype import SRV, RdataType
from ldap3 import Connection, Server, DSA
from typing import List, Callable

from ms_active_directory.environment.discovery.discovery_constants import (
    DNS_TIMEOUT_SECONDS,
    KERBEROS_DNS_SRV_FORMAT,
    KERBEROS_SITE_AWARE_DNS_SRV_FORMAT,
    LDAP_DNS_SRV_FORMAT,
    LDAP_SITE_AWARE_DNS_SRV_FORMAT,
)
from ms_active_directory import logging_utils


logger = logging_utils.get_logger()


def discover_ldap_domain_controllers_in_domain(domain: str, site: str=None, dns_nameservers: List[str]=None,
                                               source_ip: str=None, server_limit: int=None, secure: bool=True):
    """ Take in an AD domain and discover the LDAP servers in the domain that are domain
    controllers. Then order them to try and optimize going to the closest/fastest/highest priority
    servers first.
    :param domain: The string dns name of the domain.
    :param site: The string name of a site within the domain. If specified, only controllers within
                 the site will be returned.
    :param dns_nameservers: The nameservers to use for DNS lookups to discover LDAP servers. If not
                            specified, the system DNS nameservers will be used.
    :param source_ip: The source IP to use for DNS queries and when checking RTT to LDAP servers.
    :param server_limit: An integer limit on the number of controllers returned. After sorting by RTT,
                         if specified, only the fastest server_limit controllers will be returned.
    :param secure: If true, only controllers capable of securing LDAP communication using StartTLS
                   will be returned, and TLS negotiation time will be accounted for in the RTT evaluation.
    :returns: a tuple consisting of an ordered list of the fastest LDAP uris (which combines the hostname and
              protocol), and the full ordered list of LDAP uris (which combines the hostname and protocol).
    """
    logger.info('Discovering LDAP servers for domain %s in DNS', domain)
    ldap_srv = LDAP_DNS_SRV_FORMAT.format(domain=domain)
    if site:
        ldap_srv = LDAP_SITE_AWARE_DNS_SRV_FORMAT.format(site=site, domain=domain)
    all_ldap_records = _resolve_record_in_dns(ldap_srv, SRV, dns_nameservers, source_ip)
    return _order_ldap_servers_by_rtt(all_ldap_records, server_limit, source_ip, secure)


def discover_kdc_domain_controllers_in_domain(domain: str, site: str=None, dns_nameservers: List[str]=None,
                                              source_ip: str=None, server_limit: int=None):
    """ Take in an AD domain and discover the KDCs in the domain that are domain controllers. Then
    order them to try and optimize going to the closest/fastest/highest priority servers first.
    :param domain: The string dns name of the domain.
    :param site: The string name of a site within the domain. If specified, only controllers within
                 the site will be returned.
    :param dns_nameservers: The nameservers to use for DNS lookups to discover LDAP servers. If not
                            specified, the system DNS nameservers will be used.
    :param source_ip: The source IP to use for DNS queries and when checking RTT to kerberos servers.
    :param server_limit: An integer limit on the number of controllers returned. After sorting by RTT,
                         if specified, only the fastest server_limit controllers will be returned.
    :returns: a tuple consisting of a truncated list of the ordered list of KDC uris (which combines the hostname
              and port to reach out on), and the full list of KDC uris (which combines the hostname and port to
              reach out on).
    """
    logger.info('Discovering Kerberos servers for domain %s in DNS', domain)
    krb_srv = KERBEROS_DNS_SRV_FORMAT.format(domain=domain)
    if site:
        krb_srv = KERBEROS_SITE_AWARE_DNS_SRV_FORMAT.format(site=site, domain=domain)
    all_ldap_records = _resolve_record_in_dns(krb_srv, SRV, dns_nameservers, source_ip)
    return _order_kdcs_by_rtt(all_ldap_records, server_limit, source_ip)


def _resolve_record_in_dns(record_name: str, record_type: RdataType, dns_nameservers: List[str], source_ip: str):
    """ Take a record and record type and resolve it in DNS.

    Returns a list of tuples where each tuple is in the format (host, port, priority, weight)
    sorted by priority and then weight.
    """
    temp_resolver = dns.resolver.Resolver()
    temp_resolver.timeout = DNS_TIMEOUT_SECONDS
    temp_resolver.lifetime = DNS_TIMEOUT_SECONDS
    if dns_nameservers:
        logger.debug('Using the following nameservers for dns lookup instead of the default system ones %s',
                     dns_nameservers)
        temp_resolver.nameservers = dns_nameservers
    # DNS queries are normally UDP. However, the best practices from microsoft for DNS are that
    # you use TCP if your result will be greater than 512 bytes. It states that DNS may truncate
    # results greater than 512 bytes.
    # If a record maps to a lot of results (like a service record for a large domain) then our
    # result can easily exceed 512 bytes, so we use tcp directly for lookups here, rather than
    # wait for udp to fail and then fallback.
    try:
        resolved_records = temp_resolver.resolve(record_name, record_type, tcp=True,
                                                 source=source_ip)
    except dns.exception.DNSException as dns_ex:
        logger.info('Unable to query DNS for record %s due to: %s', record_name, dns_ex)
        return []
    except Exception as ex:
        logger.warning('Unexpected exception occurred when querying DNS for record %s: %s',
                       record_name, ex)
        return []

    # turn our DNS records into more manageable tuples in the form:
    # (URI, Port, Priority, Weight)
    record_tuples = [(record.target.to_text(omit_final_dot=True), record.port, record.priority, record.weight)
                     for record in resolved_records]
    # A lower priority value (closer to 0) means that a record should be preferred.
    # Weight is used to rank order records of equal priority, and a higher value weight (further
    # above 0) means that a record should be preferred.
    # So we sort ascending, first by priority, and then by -1 * weight
    record_tuples = sorted(record_tuples, key=lambda record_tuple: (record_tuple[2], -1*record_tuple[3]))
    logger.debug('Records returned in %s lookup for %s ordered by priority and weight: %s',
                 record_type, record_name, record_tuples)
    return record_tuples


def _order_ldap_servers_by_rtt(ldap_server_records: List[tuple], server_limit: int, source_ip: str, secure: bool):
    """ Take in a list of LDAP server records and determine the reachability and round trip time to each.
    Order them by RTT, fastest first, and drop unreachable servers. If there's more than our limit,
    trim the length based on our limit.
    """
    lookup_rtt_fns = []
    for server_tuple in ldap_server_records:
        # we don't care about weight and priority anymore if ordering by RTT
        server_host, server_port, _, _ = server_tuple
        fn = lambda: _check_ldap_server_availability_and_rtt(server_host, server_port, source_ip, secure)
        lookup_rtt_fns.append(fn)
    return _process_sort_return_rtt_ordering_results(lookup_rtt_fns, 'LDAP', server_limit)


def _order_kdcs_by_rtt(kdc_server_records: List[tuple], server_limit: int, source_ip: str):
    """ Take in a list of KDC server records and determine the reachability and round trip time to each.
    Order them by RTT, fastest first, and drop unreachable servers. If there's more than our limit,
    trim the length based on our limit.
    """
    lookup_rtt_fns = []
    for server_tuple in kdc_server_records:
        # we don't care about weight and priority anymore if ordering by RTT
        server_host, server_port, _, _ = server_tuple
        fn = lambda: _check_kdc_availability_and_rtt(server_host, source_ip, server_port)
        lookup_rtt_fns.append(fn)
    return _process_sort_return_rtt_ordering_results(lookup_rtt_fns, 'KDC', server_limit)


def _process_sort_return_rtt_ordering_results(lookup_rtt_fns: List[Callable], uri_desc: str,
                                              maximum_result_list_length: int):
    """ Take in a list of lookup functions that measure round trip time to the server,
    and execute all of those coroutines in parallel.

    Each coroutine returns a tuple of (round trip time, composed URI). We then filter our any
    URIs that could not be reached within our connection timeout and sort by round trip time.
    We then return the URIs in sorted order.
    """
    logger.info('Sorting %s %s servers by round trip time and removing unreachable servers',
                len(lookup_rtt_fns), uri_desc)
    rtt_results = []
    with ThreadPoolExecutor() as executor:
        running_tasks = [executor.submit(lookup_fn) for lookup_fn in lookup_rtt_fns]
        for task in running_tasks:
            rtt_results.append(task.result())
    # Drop results that are None values, as they indicate unreachable servers.
    rtt_uri_tuples = [result_tuple for result_tuple in rtt_results
                      if result_tuple is not None and result_tuple[0] is not None]
    # this will sort ascending by the first key by default, which is the round trip time. so it
    # will end up sorting the closest servers first
    rtt_uri_tuples = sorted(rtt_uri_tuples)
    logger.debug('%s servers sorted by round trip time to them: %s', uri_desc, rtt_uri_tuples)

    result_list = [uri for _, uri in rtt_uri_tuples]
    short_list = result_list
    # if we have a limit on the number of servers we want, return it
    if maximum_result_list_length and len(result_list) > maximum_result_list_length:
        short_list = result_list[:maximum_result_list_length]
        logger.info('Trimming list of %s servers to the fastest %s to reply due to total number exceeding our limit. '
                    'Remaining servers: %s', uri_desc, maximum_result_list_length, short_list)
    return short_list


def _check_ldap_server_availability_and_rtt(server_host: str, server_port: str, source_ip: str, secure: bool):
    """ Even if an LDAP server is registered in DNS, it might not be reachable for us. DNS is
    centralized, but data centers may have multiple network partitions, and there may be firewalls
    or air gaps in our way (we've seen this at customer sites).
    A server could also be down.

    This will check if a server is available and return a tuple of the URI and the time it took to
    make that check.
    This isn't exactly letting us know how quickly we can query the server, because its internal
    processing power for heavyweight queries might not be directly related to how quickly it
    responds to availability checks. But it's a decent and quick approximation.

    If secure is True, then we'll start TLS on any LDAP connection to ensure that we can negotiate
    TLS with the server. Otherwise, we'll just use plaintext LDAP.

    Returns None for any unreachable servers.
    """
    ldap_uri = 'ldap://{}:{}'.format(server_host, server_port)
    # by using get_info=DSA we perform what's sometimes referred to as form of "LDAP ping".
    # we query the root DSE, which can be done without binding
    server = Server(ldap_uri, get_info=DSA)
    conn = Connection(server, source_address=source_ip)
    start = time.time()
    try:
        conn.open()
        if secure:
            if not conn.start_tls():
                logger.debug('LDAP server %s was reachable on port %s but failed to start secure communication')
                return None
        end = time.time()
    except:
        if secure:
            logger.debug('LDAP server %s was unreachable on port %s or raised an exception establishing secure communication on that port',
                         server_host, server_port)
        else:
            logger.debug('LDAP server %s was unreachable on port %s', server_host, server_port)
        return None
    return end - start, ldap_uri


def _check_kdc_availability_and_rtt(server_host: str, server_port: str, source_ip: str):
    """ Even if an KDC server is registered in DNS, it might not be reachable for us. DNS is
    centralized, but data centers may have multiple network partitions, and there may be firewalls
    or air gaps in our way (we've seen this at customer sites).
    A server could also be down.

    This will check if a server is available and return a tuple of the URI and the time it took to
    make that check.
    This isn't exactly letting us know how quickly we can query the server, because its internal
    processing power for heavyweight queries might not be directly related to how quickly it
    responds to availability checks. But it's a decent and quick approximation.

    Returns None if the server is unreachable.
    """
    candidates = []
    source_ip = source_ip if source_ip else ''  # socket library wants an empty string for dynamic assignment
    # hostname - try both ipv4 and ipv6
    kdc_uri = format_utils.format_hostname_or_ip_and_port_to_uri(server_host, server_port,
                                                                 is_ipv6_fmt=False)

    test_socket4 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    addr_tuple4 = (server_host, server_port)

    test_socket6 = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
    addr_tuple6 = (server_host, server_port, 0, 0)

    candidates.append((test_socket6, addr_tuple6))
    candidates.append((test_socket4, addr_tuple4))

    for temp_socket, temp_addr_tuple in candidates:
        try:
            try:
                # bind before time evaluation so that the underlying network stack isn't factored in,
                # since dynamic port assignment time varies
                temp_socket.bind((source_ip, 0))  # 0 for dynamically assigned port
                start = time.time()
                temp_socket.connect(temp_addr_tuple)
                end = time.time()
                return end - start, kdc_uri
            except socket.error:
                logger.debug('KDC server %s was unreachable on port %s', server_host, server_port)
                pass
            finally:
                try:
                    temp_socket.shutdown(socket.SHUT_RDWR)
                    temp_socket.close()
                except socket.error:
                    pass
        except:
            pass
    return None




from ms_active_directory.environment.ldap.ldap_constants import (
    AD_ATTRIBUTE_COMMON_NAME,
    AD_ATTRIBUTE_OBJECT_CLASS,
    AD_ATTRIBUTE_SAMACCOUNT_NAME
)
from ms_active_directory.environment.ldap.ldap_format_utils import (
    normalize_object_location_in_domain
)


class ADUser:

    def __init__(self, dn: str, attributes: dict, domain):
        self.distinguished_name = dn
        self.domain = domain
        self.samaccount_name = attributes.get(AD_ATTRIBUTE_SAMACCOUNT_NAME)
        self.common_name = attributes.get(AD_ATTRIBUTE_COMMON_NAME)
        self.object_classes = attributes.get(AD_ATTRIBUTE_OBJECT_CLASS)
        self.name = self.samaccount_name
        # to get the location of a user, remove their common name from the front and the domain's
        # domain components from the end
        rdn_of_user = normalize_object_location_in_domain(dn, self.domain.get_domain_dns_name())
        user_piece = 'CN=' + self.common_name
        self.location = rdn_of_user[len(user_piece)+1:]
        self.other_attributes = attributes

    def get(self, attribute_name: str):
        """ Get an attribute about the group that isn't explicitly tracked as a member """
        val = self.other_attributes.get(attribute_name)
        # there's a lot of 1-item lists from the ldap3 library
        if isinstance(val, list) and len(val) == 1:
            return val[0]
        return val


class ADGroup:

    def __init__(self, dn: str, attributes: dict, domain):
        self.distinguished_name = dn
        self.domain = domain
        self.samaccount_name = attributes.get(AD_ATTRIBUTE_SAMACCOUNT_NAME)
        self.common_name = attributes.get(AD_ATTRIBUTE_COMMON_NAME)
        self.object_classes = attributes.get(AD_ATTRIBUTE_OBJECT_CLASS)
        self.name = self.samaccount_name
        # to get the location of a user, remove their common name from the front and the domain's
        # domain components from the end
        rdn_of_user = normalize_object_location_in_domain(dn, self.domain.get_domain_dns_name())
        user_piece = 'CN=' + self.common_name
        self.location = rdn_of_user[len(user_piece)+1:]
        self.other_attributes = attributes

    def get(self, attribute_name: str):
        """ Get an attribute about the group that isn't explicitly tracked as a member """
        val = self.other_attributes.get(attribute_name)
        # there's a lot of 1-item lists from the ldap3 library
        if isinstance(val, list) and len(val) == 1:
            return val[0]
        return val



from ms_active_directory import logging_utils

from typing import List

from ms_active_directory.core.ad_kerberos_keys import (
    GssKerberosKey,
)
from ms_active_directory.environment.kerberos.kerberos_constants import (
    AD_DEFAULT_NAME_TYPE
)
from ms_active_directory.environment.kerberos.kerberos_keytab_generator import (
    write_gss_kerberos_key_list_to_raw_bytes
)
from ms_active_directory.environment.kerberos.kerberos_raw_key_generator import (
    ad_password_string_to_key
)
from ms_active_directory.environment.ldap.ldap_format_utils import (
    construct_object_distinguished_name
)
from ms_active_directory.environment.security.security_config_constants import (
    ADEncryptionType,
    ENCRYPTION_TYPE_STR_TO_ENUM,
)
from ms_active_directory.exceptions import InvalidComputerParameterException

logger = logging_utils.get_logger()


class ADComputer:

    def __init__(self, samaccount_name: str, domain, location: str=None,
                 password: str=None, service_principal_names: List[str]=None,
                 encryption_types: List[ADEncryptionType]=None, kvno: int=None):
        self.samaccount_name = samaccount_name
        self.computer_name = self.samaccount_name[:-1]
        self.name = self.computer_name
        self.domain = domain
        self.domain_dns_name = self.domain.get_domain_dns_name()
        self.realm = self.domain_dns_name.upper()
        self.location = location
        self.password = password
        self.service_principal_names = service_principal_names if service_principal_names else []
        self.encryption_types = []
        encryption_types = encryption_types if encryption_types else []
        for enc_type in encryption_types:
            original = enc_type
            if isinstance(enc_type, str):
                enc_type = ENCRYPTION_TYPE_STR_TO_ENUM.get(enc_type.lower())
            if not isinstance(enc_type, ADEncryptionType):
                raise ValueError('All specified encryption types must be ADEncryptionType enums or must '
                                 'be strings convertible to ADEncryptionType enums. {} is neither.'
                                 .format(original))
            self.encryption_types.append(enc_type)
        # assume the account is a new account unless given a kvno
        self.kvno = 1 if kvno is None else self.kvno

        self.kerberos_keys = []
        self.raw_kerberos_keys = []
        self.server_kerberos_keys = []
        self.user_kerberos_keys = []
        if self.password:
            logger.debug('Generating kerberos keys from password during instantiation of computer with name %s',
                         self.computer_name)
            for enc_type in self.encryption_types:
                raw_key = ad_password_string_to_key(enc_type, self.computer_name,
                                                    self.password, self.domain_dns_name)
                self.raw_kerberos_keys.append(raw_key)

                for spn in self.service_principal_names:
                    server_gss_kerberos_key = GssKerberosKey(spn, self.realm, raw_key, self.kvno,
                                                             gss_name_type=AD_DEFAULT_NAME_TYPE)
                    self.kerberos_keys.append(server_gss_kerberos_key)
                    self.server_kerberos_keys.append(server_gss_kerberos_key)
                # generate our user kerberos key
                user_gss_kerberos_key = GssKerberosKey(self.samaccount_name, self.realm, raw_key, self.kvno,
                                                       gss_name_type=AD_DEFAULT_NAME_TYPE)
                self.kerberos_keys.append(user_gss_kerberos_key)
                self.user_kerberos_keys.append(user_gss_kerberos_key)
            logger.debug('Generated %s kerberos keys from password during instantiation of computer with name %s',
                         len(self.kerberos_keys), self.computer_name)

    def add_encryption_type_locally(self, encryption_type: ADEncryptionType):
        """ Adds an encryption type to the computer locally. This will generate new kerberos keys
        for the computer as a user and for all of the computer's service principal names using the
        new encryption type.
        This function does nothing if the encryption type is already on the computer.
        This function raises an exception if the computer's password is not set, as the password is
        needed to generate new kerberos keys.
        :param encryption_type: The encryption type to add to the computer.
        """
        if encryption_type in self.encryption_types:
            logger.debug('No change resulted from adding encryption type %s to computer %s locally as it was already present',
                         encryption_type, self.computer_name)
            return
        if self.password is None:
            raise InvalidComputerParameterException('Encryption types can only be added to a computer locally if its '
                                                    'password is known. Without the password, new kerberos keys cannot '
                                                    'be generated.')
        logger.debug('Adding encryption type %s to computer %s locally',
                     encryption_type, self.computer_name)
        self.encryption_types.append(encryption_type)
        raw_krb_key = ad_password_string_to_key(encryption_type, self.computer_name,
                                                self.password, self.domain_dns_name)
        self.raw_kerberos_keys.append(raw_krb_key)
        for spn in self.service_principal_names:
            server_gss_kerberos_key = GssKerberosKey(spn, self.realm, raw_krb_key, self.kvno,
                                                     gss_name_type=AD_DEFAULT_NAME_TYPE)
            self.kerberos_keys.append(server_gss_kerberos_key)
            self.server_kerberos_keys.append(server_gss_kerberos_key)
        user_gss_kerberos_key = GssKerberosKey(self.samaccount_name, self.realm, raw_krb_key, self.kvno,
                                               gss_name_type=AD_DEFAULT_NAME_TYPE)
        self.kerberos_keys.append(user_gss_kerberos_key)
        self.user_kerberos_keys.append(user_gss_kerberos_key)

    def add_service_principal_name_locally(self, service_principal_name: str):
        """ Adds a service principal name to the computer locally. This will generate new kerberos keys
        for the computer to use to accept security contexts for the service principal name using
        all raw kerberos keys that the account has (and therefore all encryption types it has).
        This function does nothing if the service principal name is already on the computer.

        :param service_principal_name: The service principal name to add to the computer.
        """
        if service_principal_name in self.service_principal_names:
            logger.debug('No change resulted from adding service principal name %s as it was already present for computer %s',
                         service_principal_name, self.computer_name)
            return
        logger.debug('Adding service principal name %s to computer %s locally',
                     service_principal_name, self.computer_name)
        self.service_principal_names.append(service_principal_name)
        for raw_krb_key in self.raw_kerberos_keys:
            gss_kerberos_key = GssKerberosKey(service_principal_name, self.realm, raw_krb_key, self.kvno,
                                              gss_name_type=AD_DEFAULT_NAME_TYPE)
            self.kerberos_keys.append(gss_kerberos_key)
            self.server_kerberos_keys.append(gss_kerberos_key)

    def get_full_keytab_file_bytes_for_computer(self):
        """ Get the raw bytes that would comprise a complete keytab file for this computer. The
        resultant bytes form a file that can be used to either accept GSS security contexts as a
        server for any protocol and hostname combinations defined in the service principal names,
        or initiate them as the computer with the computer's user principal name being the
        sAMAccountName.
        """
        return write_gss_kerberos_key_list_to_raw_bytes(self.kerberos_keys)

    def get_server_keytab_file_bytes_for_computer(self):
        """ Get the raw bytes that would comprise a server keytab file for this computer. The resultant
        bytes form a file that can be used to accept GSS security contexts as a server for any protocol
        and hostname combinations defined in the service principal names.
        """
        return write_gss_kerberos_key_list_to_raw_bytes(self.server_kerberos_keys)

    def get_user_keytab_file_bytes_for_computer(self):
        """ Get the raw bytes that would comprise a server keytab file for this computer. The
        resultant bytes form a file that can be used to initiate GSS security contexts as the
        computer with the computer's user principal name being the sAMAccountName.
        """
        return write_gss_kerberos_key_list_to_raw_bytes(self.user_kerberos_keys)

    def get_computer_name(self):
        return self.computer_name

    def get_computer_distinguished_name(self):
        """ Get the LDAP distinguished name for the computer. This raises an exception if location is not
        set for the computer.
        """
        if self.location is None:
            raise InvalidComputerParameterException('The location of the computer is unknown and so a distinguished '
                                                    'name cannot be determined for it.')
        return construct_object_distinguished_name(self.computer_name, self.location, self.domain_dns_name)

    def get_domain(self):
        return self.domain

    def get_domain_dns_name(self):
        return self.domain_dns_name

    def get_encryption_types(self):
        return self.encryption_types

    def get_name(self):
        return self.name

    def get_samaccount_name(self):
        return self.samaccount_name

    def get_server_kerberos_keys(self):
        return self.server_kerberos_keys

    def get_service_principal_names(self):
        return self.service_principal_names

    def get_user_kerberos_keys(self):
        return self.user_kerberos_keys

    def get_user_principal_name(self):
        """ Gets the user principal name for the computer, to be used in initiating GSS security contexts """
        return '{sam}@{realm}'.format(sam=self.samaccount_name, realm=self.realm)

    def set_encryption_types_locally(self, encryption_types: List[ADEncryptionType]):
        """ Sets the encryption types of the computer locally. This will generate new kerberos keys
        for the computer as a user and for all of the computer's service principal names using the
        new encryption type.
        This function raises an exception if the computer's password is not set, as the password is
        needed to generate new kerberos keys.
        :param encryption_types: The list of AD encryption types to set on the computer.
        """
        if self.password is None:
            raise InvalidComputerParameterException('Encryption types can only be set on a computer locally if its '
                                                    'password is known. Without the password, new kerberos keys cannot '
                                                    'be generated.')

        new_kerberos_keys = []
        new_raw_kerberos_keys = []
        new_server_kerberos_keys = []
        new_user_kerberos_keys = []
        logger.debug('Adding new encryption types %s to computer %s locally',
                     encryption_types, self.computer_name)
        for encryption_type in encryption_types:
            raw_krb_key = ad_password_string_to_key(encryption_type, self.computer_name,
                                                    self.password, self.domain_dns_name)
            new_raw_kerberos_keys.append(raw_krb_key)
            for spn in self.service_principal_names:
                server_gss_kerberos_key = GssKerberosKey(spn, self.realm, raw_krb_key, self.kvno,
                                                         gss_name_type=AD_DEFAULT_NAME_TYPE)
                new_kerberos_keys.append(server_gss_kerberos_key)
                new_server_kerberos_keys.append(server_gss_kerberos_key)
            user_gss_kerberos_key = GssKerberosKey(self.samaccount_name, self.realm, raw_krb_key, self.kvno,
                                                   gss_name_type=AD_DEFAULT_NAME_TYPE)
            new_kerberos_keys.append(user_gss_kerberos_key)
            new_user_kerberos_keys.append(user_gss_kerberos_key)
        self.encryption_types = encryption_types
        self.kerberos_keys = new_kerberos_keys
        self.raw_kerberos_keys = new_raw_kerberos_keys
        self.server_kerberos_keys = new_server_kerberos_keys
        self.user_kerberos_keys = new_user_kerberos_keys
        logger.debug('Generated %s new kerberos keys for new encryption types %s set on computer %s locally',
                     len(new_kerberos_keys), encryption_types, self.computer_name)

    def set_password_locally(self, password: str):
        """ Sets the password on the AD computer locally. This will regenerate server and user kerberos
        keys for all of the encryption types on the computer.
        This function is meant to be used when the password was not set locally or was incorrectly set.
        This function WILL NOT update the key version number of the kerberos keys; if a computer's
        password is actually changed, then update_password_locally should be used as that will update
        the key version number properly and ensure the resultant kerberos keys can be properly used
        for initiating and accepting security contexts.
        :param password: The string password to set for the computer.
        """
        self.password = password
        self.kerberos_keys = []
        self.raw_kerberos_keys = []
        self.server_kerberos_keys = []
        self.user_kerberos_keys = []
        logger.debug('Generating new kerberos keys for computer %s based on new password',
                     self.computer_name)
        for enc_type in self.encryption_types:
            raw_key = ad_password_string_to_key(enc_type, self.computer_name,
                                                self.password, self.domain_dns_name)
            self.raw_kerberos_keys.append(raw_key)

            for spn in self.service_principal_names:
                server_gss_kerberos_key = GssKerberosKey(spn, self.realm, raw_key, self.kvno,
                                                         gss_name_type=AD_DEFAULT_NAME_TYPE)
                self.kerberos_keys.append(server_gss_kerberos_key)
                self.server_kerberos_keys.append(server_gss_kerberos_key)
            # generate our user kerberos key
            user_gss_kerberos_key = GssKerberosKey(self.samaccount_name, self.realm, raw_key, self.kvno,
                                                   gss_name_type=AD_DEFAULT_NAME_TYPE)
            self.kerberos_keys.append(user_gss_kerberos_key)
            self.user_kerberos_keys.append(user_gss_kerberos_key)
        logger.info('Generated %s new kerberos keys for computer %s based on new password and forgot old keys',
                    len(self.kerberos_keys), self.computer_name)

    def set_service_principal_names_locally(self, service_principal_names: List[str]):
        """ Sets the service principal names for the computer, and regenerates new server kerberos keys
        for all of the newly set service principal names.
        :param service_principal_names: A list of string service principal names to set for the computer.
        """
        logger.debug('Generating new kerberos keys for service principal names %s set on computer %s',
                     service_principal_names, self.computer_name)
        new_kerberos_keys = []
        for spn in self.service_principal_names:
            for raw_krb_key in self.raw_kerberos_keys:
                gss_kerberos_key = GssKerberosKey(spn, self.realm, raw_krb_key, self.kvno,
                                                  gss_name_type=AD_DEFAULT_NAME_TYPE)
                new_kerberos_keys.append(gss_kerberos_key)
        self.service_principal_names = service_principal_names
        self.kerberos_keys = new_kerberos_keys + self.user_kerberos_keys
        logger.debug('Generated %s new kerberos keys for new service principal names %s set on computer %s locally',
                     len(new_kerberos_keys), service_principal_names, self.computer_name)

    def write_full_keytab_file_for_computer(self, file_path):
        logger.debug('Writing full key file for computer %s to %s', self.computer_name, file_path)
        data = self.get_full_keytab_file_bytes_for_computer()
        self._write_keytab_data(file_path, data)
        logger.info('Successfully wrote full key file with %s keys for computer %s to %s',
                    len(self.kerberos_keys), self.computer_name, file_path)

    def write_server_keytab_file_for_computer(self, file_path):
        logger.debug('Writing server key file for computer %s to %s', self.computer_name, file_path)
        data = self.get_server_keytab_file_bytes_for_computer()
        self._write_keytab_data(file_path, data)
        logger.info('Successfully wrote server key file with %s keys for computer %s to %s',
                    len(self.server_kerberos_keys), self.computer_name, file_path)

    def write_user_keytab_file_for_computer(self, file_path):
        logger.debug('Writing user key file for computer %s to %s', self.computer_name, file_path)
        data = self.get_user_keytab_file_bytes_for_computer()
        self._write_keytab_data(file_path, data)
        logger.info('Successfully wrote user key file with %s keys for computer %s to %s',
                    len(self.user_kerberos_keys), self.computer_name, file_path)

    def _write_keytab_data(self, file_path, data):
        with open(file_path, 'wb') as fp:
            fp.write(data)

    def update_password_locally(self, password: str):
        """ Update the password for the computer locally and generate new kerberos keys for the new
        password.
        :param password: The string password to set for the computer.
        """
        self.kvno += 1
        logger.debug('Updated kvno for computer %s from %s to %s', self.computer_name, self.kvno-1, self.kvno)
        self.set_password_locally(password)



import ipaddress

from ms_active_directory import logging_utils

from ldap3 import (
    KERBEROS,
    NTLM,
)

from ldap3.utils.dn import (
    parse_dn,
)

from ms_active_directory.environment.ldap.ldap_format_utils import is_dn
from ms_active_directory.exceptions import InvalidDomainParameterException


logger = logging_utils.get_logger()


def format_computer_name_for_authentication(computer_name: str, domain: str, authentication_mechanism: str):
    original_name = computer_name
    if is_dn(computer_name):
        parse_dn(computer_name)
        name_format = 'domain\\sAMAccountName' if authentication_mechanism == NTLM else 'sAMAccountName@domain'
        raise InvalidDomainParameterException('Computer names may not be specified as distinguished names for LDAP '
                                              'authentication. Please specify the computer name in the format {}'
                                              .format(name_format))

    adjusted_computer_name = computer_name.lower()
    # split up domain and computer name if they're already both included. @ and \ are not valid
    # sAMAccountName characters, so we can split on those safely.
    # if the name came with a domain in it, use that. it likely matches the domain passed in, but
    # it could be a subdomain in which case we don't want to tamper with it
    if '@' in adjusted_computer_name:
        adjusted_computer_name, domain = adjusted_computer_name.split('@')
    elif '\\' in adjusted_computer_name:
        domain, adjusted_computer_name = adjusted_computer_name.split('\\')

    # add a $ at the end of the sAMAccountName if it's not there, as is convention for computers
    if not adjusted_computer_name.endswith('$'):
        adjusted_computer_name = computer_name + '$'

    # put the name and domain into the right format
    if authentication_mechanism == NTLM:
        # user needs to be 'domain\computer_samaccount_name'
        adjusted_computer_name = domain.lower() + '\\' + adjusted_computer_name
    elif authentication_mechanism == KERBEROS:
        # user needs to be 'computer_samaccount_name@domain'
        adjusted_computer_name = adjusted_computer_name + '@' + domain.lower()
    logger.debug('Adjusted computer name %s to be %s for LDAP authentication using %s',
                 original_name, adjusted_computer_name, authentication_mechanism)
    return adjusted_computer_name


def format_hostname_or_ip_and_port_to_uri(host_or_ip: str, port: str, is_ipv6_fmt: bool=None):
    """ Combine what is either an ipv4 address, ipv6 address, or hostname and (optionally) a port
    into the proper format.
    """
    if port is None or port == '':
        return host_or_ip

    if is_ipv6_fmt is None:
        try:
            ipaddress.IPv6Address(host_or_ip)
            is_ipv6_fmt = True
        except ValueError:
            pass

    if is_ipv6_fmt:
        return '[{}]:{}'.format(host_or_ip, port)
    return '{}:{}'.format(host_or_ip, port)



""" Utilities for interacting with security-related aspects of an AD configuration, such as the
password and encryption types.
"""
import random

from ms_active_directory import logging_utils

from typing import List

from ms_active_directory.environment.security.security_config_constants import (
    AD_PASSWORD_CHAR_RANGE,
    ADEncryptionType,
    ENCRYPTION_TYPE_STR_TO_ENUM,
    ENCRYPTION_TYPE_VALUE_TO_ENUM,
    UNSUPPORTED_ENC_TYPES,
)
from ms_active_directory.exceptions import (
    InvalidLdapParameterException,
    LdapResponseDecodeException,
)


logger = logging_utils.get_logger()


def encode_password(password: str):
    """ Encodes a password to be set for an AD account via the LDAP protocol.
    Surrounds password in quotes and encodes with 'utf-16-le' as is required when setting the
    password of the computer account.
    """
    quoted_pw = '"' + password + '"'
    encoded_pw = quoted_pw.encode('utf-16-le')
    return encoded_pw


def generate_random_ad_password(password_length: int=120):
    """ Generates a random computer password by generating random characters in the valid range for
    AD passwords until we reach the specified length.
    :param password_length: The length of the password to generate. Defaults to 120 characters if not
                            specified.
    """
    password = ''
    for count in range(password_length):
        password += random.choice(AD_PASSWORD_CHAR_RANGE)
    return password


def get_supported_encryption_types_value(encryption_types: List[ADEncryptionType]):
    """ Calculates the number that represents the list of encryption types by adding their values
    in the bit map.
    """
    ret = 0
    # try to normalize our list just in case the caller ignored the type hint
    normalized_types = normalize_encryption_type_list(encryption_types)
    for encryption_type in normalized_types:
        ret += encryption_type.value
    return ret


def get_supported_encryption_type_enums_from_value(encryption_types_value: int):
    """ Given the numeric representation of encryption type that comes from an entry in an AD domain,
    return a list of encryption type enums.
    """
    # we want the higher numbers first so that our encryption types are ordered by strength
    sorted_enc_type_tuples = sorted(ENCRYPTION_TYPE_VALUE_TO_ENUM.items(), reverse=True)
    encryption_types = []
    for enc_value, encryption_type_enum in sorted_enc_type_tuples:
        # use a bitwise AND to check if this encryption type is part of our encryption type value
        if enc_value & encryption_types_value:
            encryption_types.append(encryption_type_enum)
            encryption_types_value -= enc_value

    # if we have a non-zero remainder, then there's some encryption type encoded that we don't
    # recognize
    if encryption_types_value != 0:
        raise LdapResponseDecodeException('Un-parseable encryption type value from AD: {}'
                                          .format(encryption_types_value))

    return encryption_types


def normalize_encryption_type_list(encryption_types: List[ADEncryptionType]):
    """ Given a list of encryption types, which may be strings or enums, normalize them to enums. """
    normalized_list = []
    valid_strings = sorted(ENCRYPTION_TYPE_STR_TO_ENUM.keys())
    for encryption_type in encryption_types:
        if isinstance(encryption_type, str):
            # cast to lowercase for looking in our dict
            encryption_type = ENCRYPTION_TYPE_STR_TO_ENUM.get(encryption_type.lower())
        if encryption_type is None or not isinstance(encryption_type, ADEncryptionType):
            raise InvalidLdapParameterException('All members of an encryption type list must by encryption type enums '
                                                'or must be strings that map to encryption type enums. Valid strings '
                                                'are: {}'.format(', '.join(valid_strings)))

        if encryption_type in UNSUPPORTED_ENC_TYPES:
            raise NotImplementedError('Support for encryption type {} has not been implemented'
                                      .format(encryption_type))
        normalized_list.append(encryption_type)
    return normalized_list
